-- /packages/intranet-budget/sql/postgres/intranet-budget-create.sql
--
-- Copyright (c) 2007-2018 ]project-open[
-- All rights reserved.
--
-- @author	frank.bergmann@project-open.com

-- BudgetItems
--
-- Each project can have any number of sub-budget


-------------------------------------------------------------
-- Budget Items
-------------------------------------------------------------


select acs_object_type__create_type (
	'im_budget_item',		-- object_type
	'Budget Item',			-- pretty_name
	'Budget Items',			-- pretty_plural
	'im_biz_object',		-- supertype
	'im_budget_items',		-- table_name
	'budget_item_id',		-- id_column
	'intranet-budget',		-- package_name
	'f',				-- abstract_p
	null,				-- type_extension_table
	'im_budget_item__name'		-- name_method
);


insert into acs_object_type_tables (object_type,table_name,id_column)
values ('im_budget_item', 'im_budget_items', 'budget_item_id');
insert into acs_object_type_tables (object_type,table_name,id_column)
values ('im_budget_item', 'acs_objects', 'object_id');

update acs_object_types set
	status_type_table = 'im_budget_items',
	status_column = 'budget_item_status_id',
	type_column = 'budget_item_type_id',
	type_category_type = 'Intranet Budget Item Type'
where object_type = 'im_budget_item';


-------------------------------------------------------------
-- Business Object URLs

insert into im_biz_object_urls (object_type, url_type, url) values (
'im_budget_item','view','/intranet-budget/new?form_mode=display&budget_item_id=');
insert into im_biz_object_urls (object_type, url_type, url) values (
'im_budget_item','edit','/intranet-budget/new?form_mode=edit&budget_item_id=');



create table im_budget_items (
	budget_item_id		integer
				constraint im_budget_items_pk
				primary key
				constraint im_budget_item_prj_fk
				references acs_objects,
	budget_item_name	text not null,
	budget_item_nr		text not null,

	-- Single "main" parent.
	budget_item_parent_id	integer
				constraint im_budget_items_parent_fk
				references im_budget_items,

	-- Type - deeply nested...
	budget_item_type_id	integer not null
				constraint im_budget_items_type_fk
				references im_categories,
	budget_item_status_id	integer not null
				constraint im_budget_items_status_fk
				references im_categories,

	-- Is there an owner? Takes the owner from parent if null.
	budget_item_owner_id	integer
				constraint im_budget_items_owner_fk
				references persons,

	-- To which project does the BI belong?
	budget_item_project_id	integer
				constraint im_budget_items_project_fk
				references im_projects,

	-- Maximum value for the the aggregated amount below this item
	budget_item_max_value	float,

	-- Notify the owner when reaching this amount
	budget_item_alarm_value	float,

	-- Value and date of last calculation
	budget_item_current_value_cache		float,
	budget_item_current_value_cache_date	timestamptz,

	sort_order		integer,
	description		text,
	note			text,

	-- Cache for CI hierarchy
	tree_sortkey		varbit,
	max_child_sortkey	varbit,


	match_cost_type_id	integer
				constraint im_budget_items_match_cost_type_fk
				references im_categories,
	match_expense_type_id	integer
				constraint im_budget_items_match_expense_type_fk
				references im_categories,
	match_material_type_id	integer
				constraint im_budget_items_match_material_type_fk
				references im_categories,
	match_material_id	integer
				constraint im_budget_items_match_material_fk
				references im_materials,
	match_task_id		integer
				constraint im_budget_items_match_task_fk
				references im_projects,
	match_user_id		integer
				constraint im_budget_items_match_user_fk
				references persons

);


create index im_budget_item_parent_id_idx on im_budget_items(budget_item_parent_id);
create index im_budget_item_treesort_idx on im_budget_items(tree_sortkey);
create index im_budget_item_project_id_idx on im_budget_items(budget_item_project_id);
create index im_budget_item_status_id_idx on im_budget_items(budget_item_status_id);
create index im_budget_item_type_id_idx on im_budget_items(budget_item_type_id);



-- Dont allow the same name for the same parent
alter table im_budget_items add
	constraint im_budget_items_name_un
	unique(budget_item_name, budget_item_parent_id);

-- Dont allow the same budget_item_nr for the same parent
alter table im_budget_items add
	constraint im_budget_items_nr_un
	unique(budget_item_nr, budget_item_parent_id);


-- This is the sortkey code
create or replace function im_budget_item_insert_tr ()
returns trigger as $$
declare
	v_max_child_sortkey	im_budget_items.max_child_sortkey%TYPE;
	v_parent_sortkey	im_budget_items.tree_sortkey%TYPE;
begin
	IF new.budget_item_parent_id is null THEN
		new.tree_sortkey := int_to_tree_key(new.budget_item_id+1000);
	ELSE
		select tree_sortkey, tree_increment_key(max_child_sortkey)
		into v_parent_sortkey, v_max_child_sortkey
		from im_budget_items
		where budget_item_id = new.budget_item_parent_id
		for update;

		update im_budget_items
		set max_child_sortkey = v_max_child_sortkey
		where budget_item_id = new.budget_item_parent_id;

		new.tree_sortkey := v_parent_sortkey || v_max_child_sortkey;
	END IF;
	new.max_child_sortkey := null;
	return new;
end;$$ language 'plpgsql';

create trigger im_budget_item_insert_tr
before insert on im_budget_items
for each row
execute procedure im_budget_item_insert_tr();



create or replace function im_budget_items_update_tr () 
returns trigger as $$
declare
	v_parent_sk	varbit default null;
	v_max_child_sortkey	varbit;
	v_old_parent_length	integer;
begin
	IF new.budget_item_id = old.budget_item_id
		and ((new.budget_item_parent_id = old.budget_item_parent_id)
		or (new.budget_item_parent_id is null and old.budget_item_parent_id is null)) THEN
		return new;
	END IF;

	-- the tree sortkey is going to change so get the new one and update it and all its
	-- children to have the new prefix...
	v_old_parent_length := length(new.tree_sortkey) + 1;

	IF new.budget_item_parent_id is null THEN
		v_parent_sk := int_to_tree_key(new.budget_item_id+1000);
	ELSE
		SELECT tree_sortkey, tree_increment_key(max_child_sortkey)
		INTO v_parent_sk, v_max_child_sortkey
		FROM im_budget_items
		WHERE budget_item_id = new.budget_item_parent_id
		FOR UPDATE;

		UPDATE im_budget_items
		SET max_child_sortkey = v_max_child_sortkey
		WHERE budget_item_id = new.budget_item_parent_id;

		v_parent_sk := v_parent_sk || v_max_child_sortkey;
	END IF;

	UPDATE im_budget_items
	SET tree_sortkey = v_parent_sk || substring(tree_sortkey, v_old_parent_length)
	WHERE tree_sortkey between new.tree_sortkey and tree_right(new.tree_sortkey);

	return new;
end;$$ language 'plpgsql';

create trigger im_budget_items_update_tr after update
on im_budget_items
for each row
execute procedure im_budget_items_update_tr ();



-- ------------------------------------------------------------
-- Classify cost items into budget items
-- ------------------------------------------------------------


-- Specify for each cost item to which budget item it belongs
alter table im_costs 
add column budget_item_id integer
constraint im_costs_budget_item_fk references im_budget_items;

alter table im_invoice_items
add column budget_item_id integer
constraint im_invoice_items_budget_item_fk references im_budget_items;


-- Include the procedure to determine a budget_item for a new cost_item
\i intranet-budget-finder.sql




-- ------------------------------------------------------------
-- Budget Item Constructor/Destructor
-- ------------------------------------------------------------

create or replace function im_budget_item__new (
	integer, varchar, timestamptz, integer, varchar, integer,
	varchar, varchar, integer, integer, integer, integer
) returns integer as $$
DECLARE
	p_budget_item_id		alias for $1;
	p_object_type		alias for $2;
	p_creation_date		alias for $3;
	p_creation_user		alias for $4;
	p_creation_ip		alias for $5;
	p_context_id		alias for $6;

	p_budget_item_name	alias for $7;
	p_budget_item_nr	alias for $8;
	p_budget_item_parent_id	alias for $9;
	p_budget_item_type_id	alias for $10;
	p_budget_item_status_id	alias for $11;
	p_budget_item_project_id	alias for $12;

	v_budget_item_id	integer;
BEGIN
	v_budget_item_id := acs_object__new (
		p_budget_item_id,
		p_object_type,
		p_creation_date,
		p_creation_user,
		p_creation_ip,
		p_context_id
	);
	insert into im_budget_items (
		budget_item_id, budget_item_name, budget_item_nr,
		budget_item_parent_id, budget_item_type_id, budget_item_status_id,
		budget_item_project_id
	) values (
		v_budget_item_id, p_budget_item_name, p_budget_item_nr,
		p_budget_item_parent_id, p_budget_item_type_id, p_budget_item_status_id,
		p_budget_item_project_id
	);
	return v_budget_item_id;
end;$$ language 'plpgsql';


create or replace function im_budget_item__delete (integer) 
returns integer as $$
DECLARE
	v_budget_item_id		alias for $1;
	row				RECORD;
BEGIN
	-- Delete relationships between budget item and users
	FOR row IN
		select	 	*
		from		acs_rels r,
				im_biz_object_members bom
		where		r.rel_id = bom.rel_id and 
				(r.object_id_one = v_budget_item_id OR r.object_id_two = v_budget_item_id)
	LOOP
		PERFORM im_biz_object_member__delete(row.object_id_one, row.object_id_two);
		PERFORM im_biz_object_member__delete(row.object_id_two, row.object_id_one);
	END LOOP;


	-- Delete the im_budget_items item associated with the id
	delete from 	im_budget_items
	where		budget_item_id = v_budget_item_id;

	-- Delete all the priviledges
	delete from 	acs_permissions
	where		object_id = v_budget_item_id;

	PERFORM	acs_object__delete(v_budget_item_id);

	return 0;
end;$$ language 'plpgsql';


create or replace function im_budget_item__name (integer) 
returns varchar as $$
DECLARE
	v_budget_item_id	alias for $1;
	v_name		varchar;
BEGIN
	select	budget_item_name
	into	v_name
	from	im_budget_items
	where	budget_item_id = v_budget_item_id;

	return v_name;
end;$$ language 'plpgsql';


-- Helper functions to make our queries easier to read
create or replace function im_budget_item_name_from_id (integer)
returns varchar as $$
DECLARE
	p_budget_item_id		alias for $1;
	v_budget_item_name	text;
BEGIN
	select budget_item_name
	into v_budget_item_name
	from im_budget_items
	where budget_item_id = p_budget_item_id;

	return v_budget_item_name;
end;$$ language 'plpgsql';


create or replace function im_budget_item_nr_from_id (integer)
returns varchar as $$
DECLARE
	p_budget_item_id	alias for $1;
	v_name		text;
BEGIN
	select budget_item_nr
	into v_name
	from im_budget_items
	where budget_item_id = p_budget_item_id;

	return v_name;
end;$$ language 'plpgsql';


-- Returns a space separated list of the budget_item_nr of the parents
CREATE or REPLACE FUNCTION im_budget_item_nr_parent_list(integer, varchar, integer)
RETURNS varchar as $body$
DECLARE
	p_budget_item_id		alias for $1;
	p_spacer		alias for $2;
	p_level			alias for $3;

	v_result		varchar;
	v_budget_item_nr		varchar;
	v_parent_id		integer;
BEGIN
	-- End of recursion.
	IF p_budget_item_id is NULL THEN RETURN ''; END IF;

	-- Error checking to avoid infinite loops within the DB...
	IF p_level > 20 THEN RETURN '- infinite loop with budget_item_id='||p_budget_item_id; END IF;

	-- Get the NR of the current project plus the parent_id
	select	budget_item_nr, budget_item_parent_id
	into	v_budget_item_nr, v_parent_id
	from	im_budget_items
	where	budget_item_id = p_budget_item_id;

	-- Recurse for the parent projects
	v_result = im_budget_item_nr_parent_list(v_parent_id, p_spacer, p_level+1);
	IF v_result != '' THEN v_result := v_result || p_spacer; END IF;
	v_result := v_result || v_budget_item_nr;

	RETURN v_result;
END; $body$ LANGUAGE 'plpgsql';


-- Shortcut function with only one argument
CREATE or REPLACE FUNCTION im_budget_item_nr_parent_list(integer)
RETURNS varchar as $body$
DECLARE
	p_budget_item_id		alias for $1;
BEGIN
	RETURN im_budget_item_nr_parent_list(p_budget_item_id, ' ', 0);
END; $body$ LANGUAGE 'plpgsql';





-----------------------------------------------------------
-- Full-Text Search for Budget Items
-----------------------------------------------------------

insert into im_search_object_types values (11, 'im_budget_item', 0.9);

create or replace function im_budget_items_tsearch ()
returns trigger as $$
declare
	v_string	varchar;
begin
	select	coalesce(c.budget_item_name, '') || ' ' ||
		coalesce(c.budget_item_nr, '') || ' ' ||
		coalesce(c.description, '')
	into    v_string
	from    im_budget_items c
	where   c.budget_item_id = new.budget_item_id;

	perform im_search_update(new.budget_item_id, 'im_budget_item', new.budget_item_id, v_string);

	return new;
end;$$ language 'plpgsql';


CREATE TRIGGER im_budget_items_tsearch_tr
AFTER INSERT or UPDATE ON im_budget_items
FOR EACH ROW EXECUTE PROCEDURE im_budget_items_tsearch();


-----------------------------------------------------------
-- Relationship with projects
-----------------------------------------------------------



-- --------------------------------------------------------
-- Budget Item - Relationship between projects and Budget Items
--
-- This relationship connects Projects with Budget Items


create table im_budget_item_project_rels (
	rel_id			integer
				constraint im_budget_item_project_rels_rel_fk
				references acs_rels (rel_id)
				constraint im_budget_item_project_rels_rel_pk
				primary key,
	sort_order		integer
);

select acs_rel_type__create_type (
	'im_budget_item_project_rel',	-- relationship (object) name
	'Budget Item Project Rel',	-- pretty name
	'Budget Item Project Rels',	-- pretty plural
	'relationship',			-- supertype
	'im_budget_item_project_rels',	-- table_name
	'rel_id',			-- id_column
	'intranet-budget-item-rel',	-- package_name
	'acs_object',			-- object_type_one
	'member',			-- role_one
	 0,				-- min_n_rels_one
	 null,				-- max_n_rels_one
	'im_budget_item',		-- object_type_two
	'member',			-- role_two
	0,				-- min_n_rels_two
	null				-- max_n_rels_two
);


create or replace function im_budget_item_project_rel__new (
integer, varchar, integer, integer, integer, integer, varchar, integer)
returns integer as $$
DECLARE
	p_rel_id		alias for $1;	-- null
	p_rel_type		alias for $2;	-- im_budget_item_project_rel
	p_project_id		alias for $3;
	p_budget_item_id	alias for $4;
	p_context_id		alias for $5;
	p_creation_user		alias for $6;	-- null
	p_creation_ip		alias for $7;	-- null
	p_sort_order		alias for $8;

	v_rel_id	integer;
BEGIN
	v_rel_id := acs_rel__new (
		p_rel_id,
		p_rel_type,
		p_project_id,
		p_budget_item_id,
		p_context_id,
		p_creation_user,
		p_creation_ip
	);

	insert into im_budget_item_project_rels (
	       rel_id, sort_order
	) values (
	       v_rel_id, p_sort_order
	);

	return v_rel_id;
end;$$ language 'plpgsql';


create or replace function im_budget_item_project_rel__delete (integer)
returns integer as $$
DECLARE
	p_rel_id	alias for $1;
BEGIN
	delete	from im_budget_item_project_rels
	where	rel_id = p_rel_id;

	PERFORM acs_rel__delete(p_rel_id);
	return 0;
end;$$ language 'plpgsql';


create or replace function im_budget_item_project_rel__delete (integer, integer)
returns integer as $$
DECLARE
        p_project_id	alias for $1;
	p_budget_item_id	alias for $2;

	v_rel_id	integer;
BEGIN
	select	rel_id into v_rel_id
	from	acs_rels
	where	object_id_one = p_project_id
		and object_id_two = p_budget_item_id;

	PERFORM im_budget_item_project_rel__delete(v_rel_id);
	return 0;
end;$$ language 'plpgsql';







-- ------------------------------------------------------------
-- Categories
-- ------------------------------------------------------------

-- 49000-49999  Intranet Budget (1000)
-- 49000-49099  Intranet Budget Item Status (100)
-- 49100-49199  Intranet Budget Item Type (100)
-- 49200-49299  Intranet Budget Item Action (100)
-- 49300-49399  Intranet Budget Version Status (100)
-- 49400-49499  Intranet Budget Version Type (100)


-- ---------------------------------------------------------
-- Budget Item Status
--
-- 49000-49099  Intranet Budget Item Status (100)
SELECT im_category_new(49000, 'Approved', 'Intranet Budget Item Status');
SELECT im_category_new(49010, 'Requested', 'Intranet Budget Item Status');
SELECT im_category_new(49020, 'Active', 'Intranet Budget Item Status');
SELECT im_category_new(49090, 'Closed', 'Intranet Budget Item Status');
SELECT im_category_new(49099, 'Deleted', 'Intranet Budget Item Status');
SELECT im_category_hierarchy_new(49099, 49090);

create or replace view im_budget_item_status as
select category_id as budget_item_status_id, category as budget_item_status
from im_categories
where category_type = 'Intranet Budget Item Status';


-- ---------------------------------------------------------
-- Budget Item Type
--
-- 49100-49199  Intranet Budget Item Type (100)
SELECT im_category_new(49100, 'Default', 'Intranet Budget Item Type');

create or replace view im_budget_item_type as
select category_id as budget_item_type_id, category as budget_item_type
from im_categories
where category_type = 'Intranet Budget Item Type';


SELECT im_category_new(49200, 'Save', 'Intranet Budget Item Action');
SELECT im_category_new(49290, 'Delete', 'Intranet Budget Item Action');


-----------------------------------------------------------
-- Permissions & Privileges
-----------------------------------------------------------

select acs_privilege__create_privilege('view_budget_items','View Budget Items','');
select acs_privilege__add_child('admin', 'view_budget_items');

select acs_privilege__create_privilege('view_budget_items_all','View all Budget Items','');
select acs_privilege__add_child('admin', 'view_budget_items_all');

select acs_privilege__create_privilege('edit_budget_items_all','Edit all Budget Items','');
select acs_privilege__add_child('admin', 'edit_budget_items_all');

select acs_privilege__create_privilege('add_budget_items','Add new Budget Items','');
select acs_privilege__add_child('admin', 'add_budget_items');


select im_priv_create('view_budget_items', 'Senior Managers');
select im_priv_create('view_budget_items', 'Accounting');
select im_priv_create('view_budget_items', 'Project Managers');

select im_priv_create('view_budget_items_all', 'Senior Managers');
select im_priv_create('view_budget_items_all', 'Accounting');

select im_priv_create('edit_budget_items_all', 'Senior Managers');
select im_priv_create('edit_budget_items_all', 'Accounting');

select im_priv_create('add_budget_items', 'Senior Managers');
select im_priv_create('add_budget_items', 'Accounting');
select im_priv_create('add_budget_items', 'Project Managers');



--------------------------------------------------------------
-- Business Object Roles

-- Setup the list of roles that a user can take with
-- respect to a company:
--      Full Member (1300) and
--      ToDo: Rename: Key Account Manager (1302)
--
insert into im_biz_object_role_map values ('im_budget_item',85,1300);


-- insert into im_categories (category_id, category, category_type, category_gif, category_description) values (1310, 'Budget Item Manager', 'Intranet Biz Object Role', 'member', 'Budget Item Manager');
SELECT im_category_new(1310, 'Budget Item Manager', 'Intranet Biz Object Role');
insert into im_biz_object_role_map values ('im_budget_item',85,1310);



-- -------------------------------------------------------------------
-- BudgetItem List Page Budget
-- -------------------------------------------------------------------

--
-- 310-319		intranet-budget
--
--
-- Wide View in BudgetItemListPage, including Description
--
delete from im_view_columns where view_id = 310;
delete from im_views where view_id = 310;
--
insert into im_views (view_id, view_name, visible_for) values (310, 'im_budget_item_list', 'view_budget_items');



delete from im_view_columns where column_id = 31101;
insert into im_view_columns (column_id, view_id, group_id, column_name, column_render_tcl,
extra_select, extra_where, sort_order, visible_for) values (31001,310,NULL,
'"<input type=checkbox name=_dummy onclick=acs_ListCheckAll(''budget_item'',this.checked)>"',
'"<input type=checkbox name=budget_item_id.$budget_item_id id=budget_item,$budget_item_id>"', 
'', '', 1, '');

insert into im_view_columns (column_id, view_id, group_id, column_name, column_render_tcl,
extra_select, extra_where, sort_order, visible_for) values (31005, 310, NULL, 'Name',
'"<nobr>$indent_short_html$gif_html<a href=$object_url>$budget_item_name</a></nobr>"','','',5,'');

insert into im_view_columns (column_id, view_id, group_id, column_name, column_render_tcl,
extra_select, extra_where, sort_order, visible_for) values (31007, 310, NULL, '"Max. Value"',
'"<input type=text name=budget_item_max_value.$budget_item_id size=5 value=$budget_item_max_value>"','','',7,'');

insert into im_view_columns (column_id, view_id, group_id, column_name, column_render_tcl,
extra_select, extra_where, sort_order, visible_for) values (31010, 310, NULL, 'Type',
'"<nobr>$budget_item_type</nobr>"','','',10,'');

insert into im_view_columns (column_id, view_id, group_id, column_name, column_render_tcl,
extra_select, extra_where, sort_order, visible_for) values (31020, 310, NULL, 'Status',
'"<nobr>$budget_item_status</nobr>"','','',20,'');

-- insert into im_view_columns (column_id, view_id, group_id, column_name, column_render_tcl,
-- extra_select, extra_where, sort_order, visible_for) values (31030, 310, NULL, 'Object',
-- '"<nobr>$budget_item_project_name</nobr>"','','',30,'');

insert into im_view_columns (column_id, view_id, group_id, column_name, column_render_tcl,
extra_select, extra_where, sort_order, visible_for) values (31040, 310, NULL, '"Cost Type"',
'"[im_category_select -include_empty_p 1 -include_empty_name "" {Intranet Cost Type} match_cost_type_id.$budget_item_id $match_cost_type_id]"','','',40,'');

insert into im_view_columns (column_id, view_id, group_id, column_name, column_render_tcl,
extra_select, extra_where, sort_order, visible_for) values (31050, 310, NULL, '"Expense Type"',
'"[im_category_select -include_empty_p 1 -include_empty_name "" {Intranet Expense Type} match_expense_type_id.$budget_item_id $match_expense_type_id]"','','',50,'');

insert into im_view_columns (column_id, view_id, group_id, column_name, column_render_tcl,
extra_select, extra_where, sort_order, visible_for) values (31060, 310, NULL, '"Phase"',
'"[im_select -ad_form_option_list_style_p 1 match_task_id.$budget_item_id [im_timesheet_task_options -include_empty_p 1 -main_project_id $object_id] $match_task_id]"','','',60,'');

insert into im_view_columns (column_id, view_id, group_id, column_name, column_render_tcl,
extra_select, extra_where, sort_order, visible_for) values (31070, 310, NULL, '"Material Type"',
'"[im_category_select -include_empty_p 1 -include_empty_name "" {Intranet Material Type} match_material_type_id.$budget_item_id $match_material_type_id]"','','',70,'');

insert into im_view_columns (column_id, view_id, group_id, column_name, column_render_tcl,
extra_select, extra_where, sort_order, visible_for) values (31090, 310, NULL, 'Descr',
'"$description"','','',90,'');

insert into im_view_columns (column_id, view_id, group_id, column_name, column_render_tcl,
extra_select, extra_where, sort_order, visible_for) values (31092, 310, NULL, 'Note',
'"$note"','','',92,'');




--
-- short view for project page
--
delete from im_view_columns where view_id = 311;
delete from im_views where view_id = 311;
--
insert into im_views (view_id, view_name, visible_for) values (311, 'im_budget_item_list_short', 'view_budget_items');

delete from im_view_columns where column_id = 31101;
insert into im_view_columns (column_id, view_id, group_id, column_name, column_render_tcl,
extra_select, extra_where, sort_order, visible_for) values (31101,311,NULL,
'"<input type=checkbox name=_dummy onclick=acs_ListCheckAll(''budget_item'',this.checked)>"',
'"<input type=checkbox name=budget_item_id.$budget_item_id id=budget_item,$budget_item_id>"', 
'', '', 1, '');

insert into im_view_columns (column_id, view_id, group_id, column_name, column_render_tcl,
extra_select, extra_where, sort_order, visible_for) values (31105, 311, NULL, 'Name',
'"<nobr>$indent_short_html$gif_html<a href=$object_url>$budget_item_name</a></nobr>"','','',5,'');

insert into im_view_columns (column_id, view_id, group_id, column_name, column_render_tcl,
extra_select, extra_where, sort_order, visible_for) values (31110, 311, NULL, 'Type',
'"<nobr>$budget_item_type</nobr>"','','',10,'');





-----------------------------------------------------------
-- Components & Menus
-----------------------------------------------------------

SELECT im_component_plugin__new (
	null, 'im_component_plugin', now(), null, null, null,
	'Project Budget Items',	-- plugin_name
	'intranet-budget',		-- package_name
	'right',			-- location
	'/intranet/projects/view',	-- page_url
	null,				-- view_name
	290,				-- sort_order
	'im_budget_item_list_component -object_id $project_id'	-- component_tcl
);



SELECT im_component_plugin__new (
	null, 'im_component_plugin', now(), null, null, null,
	'Project Budget Tree',	-- plugin_name
	'intranet-budget',		-- package_name
	'right',			-- location
	'/intranet/projects/view',	-- page_url
	null,				-- view_name
	300,				-- sort_order
	'im_budget_tree_component -project_id $project_id'	-- component_tcl
);




SELECT im_component_plugin__new (
	null, 'im_component_plugin', now(), null, null, null,
	'Cost Center Budget Items',	-- plugin_name
	'intranet-budget',		-- package_name
	'right',			-- location
	'/intranet-cost/cost-centers/view',		-- page_url
	null,				-- view_name
	100,				-- sort_order
	'im_budget_item_list_component -object_id $cost_center_id'	-- component_tcl
);


-----------------------------------------------------------
-- Menu for Budget Items
--

-- SELECT im_menu__new (
-- 		null, 'im_menu', now(), null, null, null,
-- 		'intranet-budget',		-- package_name
-- 		'budget_items_list',		-- label
-- 		'Budget Items List',		-- name
-- 		'/intranet-budget/index',
-- 		10,				-- sort_order
-- 		(select menu_id from im_menus where label = 'budget_items'),		-- parent_menu_id
-- 		null
-- );
-- SELECT acs_permission__grant_permission(
-- 	(select menu_id from im_menus where label = 'budget_items_list'),
-- 	(select group_id from groups where group_name = 'Employees'), 
-- 	'read'
-- );
-- 
-- 
-- 
-- SELECT im_menu__new (
-- 		null, 'im_menu', now(), null, null, null,
-- 		'intranet-budget',		-- package_name
-- 		'budget_items_dashboard',		-- label
-- 		'Budget Items Dashboard',		-- name
-- 		'/intranet-budget/dashboard',
-- 		20,				-- sort_order
-- 		(select menu_id from im_menus where label = 'budget_items'),		-- parent_menu_id
-- 		null
-- );
-- SELECT acs_permission__grant_permission(
-- 	(select menu_id from im_menus where label = 'budget_items_dashboard'),
-- 	(select group_id from groups where group_name = 'Employees'), 
-- 	'read'
-- );
-- 
-- 
-- SELECT im_menu__new (
-- 		null, 'im_menu', now(), null, null, null,
-- 		'intranet-core',		-- package_name
-- 		'budget_items_admin',		-- label
-- 		'Budget Items Admin',		-- name
-- 		'/intranet/admin/object-type-admin?object_type=im_budget_item',
-- 		900,				-- sort_order
-- 		(select menu_id from im_menus where label = 'object_type_admin'),		-- parent_menu_id
-- 		null
-- );

-- 
-- 
-- create or replace function inline_0 ()
-- returns integer as $$
-- declare
-- 	-- Menu IDs
-- 	v_menu			integer;
-- 	v_main_menu		integer;
-- 
-- 	-- Groups
-- 	v_employees		integer;
-- 	v_accounting		integer;
-- 	v_senman		integer;
-- 	v_companies		integer;
-- 	v_freelancers		integer;
-- 	v_proman		integer;
-- 	v_admins		integer;
-- 	v_reg_users		integer;
-- BEGIN
-- 	-- Get some group IDs
-- 	select group_id into v_admins from groups where group_name = 'P/O Admins';
-- 	select group_id into v_senman from groups where group_name = 'Senior Managers';
-- 	select group_id into v_proman from groups where group_name = 'Project Managers';
-- 	select group_id into v_accounting from groups where group_name = 'Accounting';
-- 	select group_id into v_employees from groups where group_name = 'Employees';
-- 	select group_id into v_companies from groups where group_name = 'Customers';
-- 	select group_id into v_freelancers from groups where group_name = 'Freelancers';
-- 	select group_id into v_reg_users from groups where group_name = 'Registered Users';
-- 
-- 	-- Determine the main menu. "Label" is used to identify menus.
-- 	select menu_id into v_main_menu 
-- 	from im_menus where label='main';
-- 
-- 	-- Create the menu.
-- 	v_menu := im_menu__new (
-- 		null,			-- p_menu_id
-- 		'im_menu',		-- object_type
-- 		now(),			-- creation_date
-- 		null,			-- creation_user
-- 		null,			-- creation_ip
-- 		null,			-- context_id
-- 		'intranet-budget',	-- package_name
-- 		'budget_items',		-- label
-- 		'Budget Items',		-- name
-- 		'/intranet-budget/index',   -- url
-- 		95,			-- sort_order
-- 		v_main_menu,		-- parent_menu_id
-- 		null			-- p_visible_tcl
-- 	);
-- 
-- 	-- Grant read permissions to most of the system
-- 	PERFORM acs_permission__grant_permission(v_menu, v_admins, 'read');
-- 	PERFORM acs_permission__grant_permission(v_menu, v_senman, 'read');
-- 	PERFORM acs_permission__grant_permission(v_menu, v_proman, 'read');
-- 	PERFORM acs_permission__grant_permission(v_menu, v_accounting, 'read');
-- 	PERFORM acs_permission__grant_permission(v_menu, v_employees, 'read');
-- 
-- 	return 0;
-- end;$$ language 'plpgsql';
-- select inline_0 ();
-- drop function inline_0 ();
-- 
-- 
-- 
-- -----------------------------------------------------------
-- -- Create new Budget Item Menu item
-- --
-- create or replace function inline_0 ()
-- returns integer as $$
-- declare
-- 	-- Menu IDs
-- 	v_menu			integer;
-- 	v_main_menu		integer;
-- 	v_employees		integer;
-- BEGIN
-- 	-- Determine the main menu. "Label" is used to identify menus.
-- 	select menu_id into v_main_menu
-- 	from im_menus where label='budget_items';
-- 
-- 	-- Create the menu.
-- 	v_menu := im_menu__new (
-- 		null,				-- p_menu_id
-- 		'im_menu',			-- object_type
-- 		now(),				-- creation_date
-- 		null,				-- creation_user
-- 		null,				-- creation_ip
-- 		null,				-- context_id
-- 		'intranet-budget',		-- package_name
-- 		'budget_items_new',		-- label
-- 		'Add a new Budget Item',	-- name
-- 		'/intranet-budget/new?form%5fmode=edit',	-- url
-- 		100,				-- sort_order
-- 		v_main_menu,			-- parent_menu_id
-- 		'im_permission $user_id add_budget_items'	-- p_visible_tcl
-- 	);
-- 
-- 	-- Grant read permissions to most of the system
-- 	select group_id into v_employees from groups where group_name = 'Employees';
-- 
-- 	return 0;
-- end;$$ language 'plpgsql';
-- select inline_0 ();
-- drop function inline_0 ();
-- 
-- 
-- 
-- 
-- -- Budget Items
-- update im_menus set sort_order = 300					where label = 'budget_items';
-- update im_menus set parent_menu_id = (select menu_id from im_menus where label = 'master_data') where label = 'budget_items';
-- update im_menus set sort_order = 10, name = 'New Budget Item'		where label = 'budget_item_add';
-- update im_menus set sort_order = 100					where label = 'budget_items_list';
-- update im_menus set sort_order = 110					where label = 'budget_items_dashboard';
-- update im_menus set sort_order = 300					where label = 'budget_item_csv_export';
-- update im_menus set sort_order = 310					where label = 'budget_item_csv_import';
-- 




-----------------------------------------------------------
-- Member component for CIs
-----------------------------------------------------------

SELECT  im_component_plugin__new (
	null, 'im_component_plugin', now(), null, null, null,
	'Budget Item Members',		-- plugin_name
	'intranet-budget',		-- package_name
	'right',			-- location
	'/intranet-budget/new',		-- page_url
	null,				-- view_name	
	20,				-- sort_order
	'im_group_member_component $budget_item_id $current_user_id $user_admin_p $return_url "" "" 1'
);

SELECT acs_permission__grant_permission(
        (select plugin_id from im_component_plugins where plugin_name = 'Budget Item Members' and package_name = 'intranet-budget'),
        (select group_id from groups where group_name = 'Employees'),
        'read'
);


-- ------------------------------------------------------
-- Show related objects
--
SELECT	im_component_plugin__new (
	null, 'im_component_plugin', now(), null, null, null,
	'Budget Item Related Objects',	-- plugin_name
	'intranet-budget',		-- package_name
	'right',			-- location
	'/intranet-budget/new',		-- page_url
	null,				-- view_name
	91,				-- sort_order
	'im_budget_item_related_objects_component -budget_item_id $budget_item_id',
	'lang::message::lookup "" intranet-budget.Budget_Item_Related_Objects "Budget Item Related Objects"'
);

SELECT acs_permission__grant_permission(
        (select plugin_id from im_component_plugins where plugin_name = 'Budget Item Related Objects' and package_name = 'intranet-budget'),
        (select group_id from groups where group_name = 'Employees'),
        'read'
);



-----------------------------------------------------------
-- DynFields
-----------------------------------------------------------

SELECT im_dynfield_widget__new (
	null, 'im_dynfield_widget', now(), 0, '0.0.0.0', null,
	'budget_item_type', 'Budget Item Type', 'Budget Item Type',
	10007, 'integer', 'im_category_tree', 'integer',
	'{custom {category_type "Intranet Budget Item Type"}}'
);
update im_dynfield_widgets
set deref_plpgsql_function = 'im_category_from_id'
where widget_name = 'budget_item_type';

SELECT im_dynfield_widget__new (
	null, 'im_dynfield_widget', now(), 0, '0.0.0.0', null,
	'budget_item_status', 'Budget Item Status', 'Budget Item Status',
	10007, 'integer', 'im_category_tree', 'integer',
	'{custom {category_type "Intranet Budget Item Status"}}'
);
update im_dynfield_widgets
set deref_plpgsql_function = 'im_category_from_id'
where widget_name = 'budget_item_status';


SELECT im_dynfield_widget__new (
	null, 'im_dynfield_widget', now(), 0, '0.0.0.0', null,
	'budget_screen_project_phase', 'Budget Screen Project Phase', 'Budget Screen Project Phase',
	10007, 'integer', 'generic_sql', 'integer',
	'{custom {sql {...}}}'
);

update im_dynfield_widgets set
	deref_plpgsql_function = 'acs_object__name',
	parameters = '{custom {sql {
select	p.project_id, p.project_name
from	im_projects p
where	p.parent_id is null
}}
after_html {
<script type="text/javascript">
function budgetScreenProjectPhaseOnChange() {
	 xmlHttp = new XMLHttpRequest();
	 xmlHttp.onreadystatechange = function() {
		if (xmlHttp.readyState == 4) {
			var phaseSelect = document.budget_item.match_task_id;
			phaseSelect.innerHTML = this.responseText;
		}
	}
	var project_id = document.budget_item.budget_item_project_id.value;
	xmlHttp.open("GET","/intranet-budget/project-phases.html?project_id="+project_id, true);
	xmlHttp.send(null);
};

	var dropdown = document.budget_item.budget_item_project_id;
	dropdown.onchange = budgetScreenProjectPhaseOnChange;
	budgetScreenProjectPhaseOnChange();

window.onload = function() {
}

</script>
}}'
where widget_name = 'budget_screen_project_phase';









SELECT im_dynfield_widget__new (
	null, 'im_dynfield_widget', now(), 0, '0.0.0.0', null,
	'budget_screen_user', 'Budget Screen User', 'Budget Screen User',
	10007, 'integer', 'generic_sql', 'integer',
	'{custom {sql {...}}}'
);

update im_dynfield_widgets set
	deref_plpgsql_function = 'acs_object__name',
	parameters = '{custom {sql {
select	p.project_id, p.project_name
from	im_projects p
where	p.parent_id is null
}}
after_html {
<script type="text/javascript">
function budgetScreenProjectPhaseOnChange() {
	 xmlHttp = new XMLHttpRequest();
	 xmlHttp.onreadystatechange = function() {
		if (xmlHttp.readyState == 4) {
			var phaseSelect = document.budget_item.match_task_id;
			phaseSelect.innerHTML = this.responseText;
		}
	}
	var project_id = document.budget_item.budget_item_project_id.value;
	xmlHttp.open("GET","/intranet-budget/project-phases.html?project_id="+project_id, true);
	xmlHttp.send(null);
};
	var dropdown = document.budget_item.budget_item_project_id;
	dropdown.onchange = budgetScreenProjectPhaseOnChange;
	budgetScreenProjectPhaseOnChange();
window.onload = function() {
}
</script>
}}'
where widget_name = 'budget_screen_user';








SELECT im_dynfield_widget__new (
	null, 'im_dynfield_widget', now(), 0, '0.0.0.0', null,
	'budget_screen_budget_item_parents', 'Budget Screen Budget Item Parents', 'Budget Screen Budget Item Parents',
	10007, 'integer', 'generic_sql', 'integer',
	'{custom {sql {...}}}'
);

update im_dynfield_widgets set
	deref_plpgsql_function = 'acs_object__name',
	parameters = '{custom {sql {
select	p.project_id, p.project_name
from	im_projects p
where	p.parent_id is null
}}
after_html {
<script type="text/javascript">
function budgetScreenProjectPhaseOnChange() {
	 xmlHttp = new XMLHttpRequest();
	 xmlHttp.onreadystatechange = function() {
		if (xmlHttp.readyState == 4) {
			var phaseSelect = document.budget_item.match_task_id;
			phaseSelect.innerHTML = this.responseText;
		}
	}
	var project_id = document.budget_item.budget_item_project_id.value;
	xmlHttp.open("GET","/intranet-budget/project-phases.html?project_id="+project_id, true);
	xmlHttp.send(null);
};

	var dropdown = document.budget_item.budget_item_project_id;
	dropdown.onchange = budgetScreenProjectPhaseOnChange;
	budgetScreenProjectPhaseOnChange();

window.onload = function() {
}

</script>
}}'
where widget_name = 'budget_screen_budget_item_parents';








SELECT im_menu__new (
	null,'im_menu',now(),null,null,null,
	'intranet-budget',					-- package_name
	'dynfield_otype_budget_item',				-- label
	'Budget Item',						-- name
	'/intranet-dynfield/object-type?object_type=im_budget_item', -- url
	112,							-- sort_order
	(select menu_id from im_menus where label = 'dynfield_otype'),	-- parent_menu_id
	null							-- p_visible_tcl
);




-----------------------------------------------------------
-- Hard Coded DynFields
--
SELECT im_dynfield_attribute_new (
	'im_budget_item', 'budget_item_name', 'Name', 
	'textbox_medium', 'string', 'f', 0, 't', 'im_budget_items'
);
SELECT im_dynfield_attribute_new (
	'im_budget_item', 'budget_item_nr', 'Nr', 
	'textbox_medium', 'string', 'f', 10, 't', 'im_budget_items'
);

SELECT im_dynfield_attribute_new (
	'im_budget_item', 'budget_item_project_id', 'Related Project',
	'open_projects', 'integer', 'f', 20, 't', 'im_budget_items'
);
SELECT im_dynfield_attribute_new (
	'im_budget_item', 'budget_item_parent_id', 'Parent', 
	'budget_screen_budget_item_parents', 'integer', 'f', 30, 't', 'im_budget_items'
);
SELECT im_dynfield_attribute_new (
	'im_budget_item', 'budget_item_type_id', 'Type', 
	'budget_item_type', 'integer', 'f', 60, 't', 'im_budget_items'
);
SELECT im_dynfield_attribute_new (
	'im_budget_item', 'budget_item_status_id', 'Status', 
	'budget_item_status', 'integer', 'f', 70, 't', 'im_budget_items'
);
SELECT im_dynfield_attribute_new (
	'im_budget_item', 'description', 'Description', 
	'textarea_small', 'string', 'f', 80, 't', 'im_budget_items'
);
SELECT im_dynfield_attribute_new (
	'im_budget_item', 'note', 'Note', 
	'textarea_small', 'string', 'f', 90, 't', 'im_budget_items'
);




-----------------------------------------------------------
-- Other fields
--

SELECT im_dynfield_attribute_new ('im_budget_item', 'match_cost_type_id', 'Cost Type', 'category_cost_type', 'integer', 'f');
insert into im_dynfield_type_attribute_map (
	attribute_id, object_type_id, display_mode, section_heading
) values (
	(select attribute_id from im_dynfield_attributes where acs_attribute_id in (
		select attribute_id from acs_attributes where attribute_name = 'match_cost_type_id'
	)),
	49100, 'edit', 'Classifiers'
);

SELECT im_dynfield_attribute_new ('im_budget_item', 'match_expense_type_id', 'Expense Type', 'category_expense_type', 'integer', 'f');
insert into im_dynfield_type_attribute_map (
	attribute_id, object_type_id, display_mode, section_heading
) values (
	(select attribute_id from im_dynfield_attributes where acs_attribute_id in (
		select attribute_id from acs_attributes where attribute_name = 'match_expense_type_id'
	)),
	49100, 'edit', 'Classifiers'
);



SELECT im_dynfield_attribute_new ('im_budget_item', 'match_task_id', 'Project Phase', 'budget_screen_project_phase', 'integer', 'f');
insert into im_dynfield_type_attribute_map (
	attribute_id, object_type_id, display_mode, section_heading
) values (
	(select attribute_id from im_dynfield_attributes where acs_attribute_id in (
		select attribute_id from acs_attributes where attribute_name = 'match_task_id'
	)),
	49100, 'edit', 'Classifiers'
);

SELECT im_dynfield_attribute_new ('im_budget_item', 'match_material_type_id', 'Material Type', 
'category_material_type', 'integer', 'f');
insert into im_dynfield_type_attribute_map (
	attribute_id, object_type_id, display_mode, section_heading
) values (
	(select attribute_id from im_dynfield_attributes where acs_attribute_id in (
		select attribute_id from acs_attributes where attribute_name = 'match_material_type_id'
	)),
	49100, 'edit', 'Classifiers'
);

SELECT im_dynfield_attribute_new ('im_budget_item', 'match_user_id', 'User', 'budget_screen_user', 'integer', 'f');
insert into im_dynfield_type_attribute_map (
	attribute_id, object_type_id, display_mode, section_heading
) values (
	(select attribute_id from im_dynfield_attributes where acs_attribute_id in (
		select attribute_id from acs_attributes where attribute_name = 'match_user_id'
	)),
	49100, 'edit', 'Classifiers'
);





\i intranet-budget-version-create.sql

