-- /packages/intranet-budget/sql/postgres/intranet-budget-finder.sql
--
-- Copyright (c) 2018 ]project-open[
-- All rights reserved.
--
-- @author	frank.bergmann@project-open.com


-- ------------------------------------------------------------
-- Classify cost items into budget items
-- ------------------------------------------------------------

DROP FUNCTION IF EXISTS im_cost_item_budget_finder_default (integer, integer, integer) CASCADE;
DROP FUNCTION IF EXISTS im_cost_item_budget_finder_invoice (integer, integer, integer) CASCADE;
DROP FUNCTION IF EXISTS im_cost_item_budget_finder_tr () CASCADE;


-- Classify a "standard" cost_item into the budget items of its project
CREATE or REPLACE function im_cost_item_budget_finder_default (integer, integer, integer)
RETURNS integer as $$
DECLARE
	p_cost_id			alias for $1;
	p_main_project_id		alias for $2;
	p_main_budget_item_id		alias for $3;

	row				record;
	v_best_match_budget_item_id	integer;

	v_cost_type_id			integer;
BEGIN
	RAISE NOTICE 'im_cost_item_budget_finder_default: cost_id=%: starting: project_id=%, main_budget_item_id=%', 
		p_cost_id, p_main_project_id, p_main_budget_item_id;

	-- Extract information from cost item
	select	cost_type_id
	into	v_cost_type_id
	from	im_costs
	where	cost_id = p_cost_id;

	-- Check for the best matching BI
	v_best_match_budget_item_id := null;
	FOR row IN
		select	t.*,
			score_cost_type + score_cost_type_subcategory as score
		from	(
			select	bi.*,
				CASE 
				WHEN v_cost_type_id = bi.match_cost_type_id THEN 1
				WHEN v_cost_type_id != bi.match_cost_type_id THEN -1000
				ELSE 0 END as score_cost_type,
				CASE
				WHEN bi.match_cost_type_id is null THEN 0
				WHEN v_cost_type_id in (select im_sub_categories(bi.match_cost_type_id)) THEN 1 
				WHEN v_cost_type_id not in (select im_sub_categories(bi.match_cost_type_id)) THEN -1000
				ELSE 0 END as score_cost_type_subcategory
			from	im_budget_items bi,
				im_budget_items parent_bi
			where	parent_bi.budget_item_id = p_main_budget_item_id and
				bi.tree_sortkey between parent_bi.tree_sortkey and tree_right(parent_bi.tree_sortkey)
			) t
		order by score
	LOOP
		RAISE NOTICE 'im_cost_item_budget_finder_default: cost_id=%: found score: bid=%, score=%', 
			p_cost_id, row.budget_item_id, row.score;

		v_best_match_budget_item_id := row.budget_item_id;
	END LOOP;

	update im_costs set budget_item_id = v_best_match_budget_item_id where cost_id = p_cost_id;
	RAISE NOTICE 'im_cost_item_budget_finder_default: cost_id=%: best match budget_item_id=%',
		p_cost_id, v_best_match_budget_item_id;

	return 0;
END;$$ language 'plpgsql';


-- Classify an "invoice" financial document (invoice, quote, purchase order, ...) 
-- into the budget items of its project.
-- Basically we have to do this will a invoice_items
CREATE or REPLACE function im_cost_item_budget_finder_invoice (integer, integer, integer)
RETURNS integer as $$
DECLARE
	p_cost_id			alias for $1;
	p_main_project_id		alias for $2;
	p_main_budget_item_id		alias for $3;

BEGIN
	RAISE NOTICE 'im_cost_item_budget_finder_invoice: cost_id=%: not implemented yet: project_id=%, main_budget_item_id=%', 
		p_cost_id, p_main_project_id, p_main_budget_item_id;

	return 0;
END;$$ language 'plpgsql';




-- This is the sortkey code
CREATE or REPLACE function im_cost_item_budget_finder_tr ()
RETURNS trigger as $$
DECLARE
	v_project_id			integer;
	v_project_name			text;
	v_project_nr			text;
	v_top_project_budget_item_id	integer;
	v_best_match_budget_item_id	integer;
	v_object_type			varchar;

	row				record;
BEGIN
	RAISE NOTICE 'im_cost_item_budget_finder_tr: cost_id=%: starting', new.cost_id;

	-------------------------------------------------------------
	-- 0. Skip if the cost has already been classified (budget_item_id already exists).
	--    Skip also customer documents (quote, delivery note and invoice).
	IF new.cost_type_id in (select im_sub_categories(3708)) THEN return new; END IF;
	IF new.budget_item_id is not null THEN 
		RAISE NOTICE 'im_cost_item_budget_finder_tr: cost_id=%: skipping, because budget_item_id=% was already set', 
			new.cost_id, new.budget_item_id;
		return new;
	END IF;

	-------------------------------------------------------------
	-- 1. Determine the project of the cost item and skip if not found.
	--
	-- Simple case - the cost is associated to one project
	select	min(main_p.project_id) into v_project_id
	from	im_projects main_p,
		im_projects p
	where	main_p.tree_sortkey = tree_root_key(p.tree_sortkey) and
		p.project_id = new.project_id;

	-- More complex - check if cost is associated to multiple projects
	IF v_project_id is null THEN
		select	min(main_p.project_id) into v_project_id
		from	im_projects p, im_projects main_p, acs_rels r
		where	r.object_id_one = new.cost_id and
			r.object_id_two = p.project_id and
			main_p.tree_sortkey = tree_root_key(p.tree_sortkey);
	END IF;

	-- Still no project? That's possible, just skip then...
	IF v_project_id is NULL THEN return new; END IF;

	-- Get the project information
	select project_name, project_nr into v_project_name, v_project_nr 
	from im_projects where project_id = v_project_id;
	RAISE NOTICE 'im_cost_item_budget_finder_tr: cost_id=%: found project: project_id=%, project_name=%', 
		new.cost_id, v_project_id, v_project_name;


	-------------------------------------------------------------
	-- 2. Make sure there is at least one budget item in the project
	--
	RAISE NOTICE 'im_cost_item_budget_finder_tr: cost_id=%: searching for top BI for project_id=%', new.cost_id, v_project_id;
	select	min(bi.budget_item_id) into v_top_project_budget_item_id
	from	im_budget_items bi
	where	bi.budget_item_project_id in (
			select	main_p.project_id
			from	im_projects main_p,
				im_projects p
			where	p.project_id = v_project_id and
				main_p.tree_sortkey = tree_root_key(p.tree_sortkey)
		) and
		bi.budget_item_parent_id is null;

	IF v_top_project_budget_item_id is null THEN
		-- Create a new top default budget item
		v_top_project_budget_item_id := im_budget_item__new(
			null, 'im_budget_item', now(), 0, '0.0.0.0', null,
			'Top Budget Item for ' || v_project_name,
			'budget_item_' || v_project_nr,
			null, 49100, 49000, v_project_id
		);
		RAISE NOTICE 'im_cost_item_budget_finder_tr: cost_id=%: created a new top budget_item_id=% for project_id=%',
			new.cost_id, v_top_project_budget_item_id, v_project_id;
	ELSE
		RAISE NOTICE 'im_cost_item_budget_finder_tr: cost_id=%: found existing top budget_item_id=% for project_id=%',
			new.cost_id, v_top_project_budget_item_id, v_project_id;
	END IF;

	
	RAISE NOTICE 'im_cost_item_budget_finder_tr: cost_id=%: before finder_default', new.cost_id;
	PERFORM im_cost_item_budget_finder_default(new.cost_id, v_project_id, v_top_project_budget_item_id);

	-------------------------------------------------------------
	-- 3. Handle different types of financial documents
	select object_type into v_object_type from acs_objects where object_id = new.cost_id;
	IF v_object_type = 'im_invoice' THEN
	   RAISE NOTICE 'im_cost_item_budget_finder_tr: cost_id=%: before finder_invoice', new.cost_id;
		PERFORM im_cost_item_budget_finder_invoice(new.cost_id, v_project_id, v_top_project_budget_item_id);
	END IF;

	return new;
END;$$ language 'plpgsql';

drop trigger if exists im_cost_item_budget_finder_tr on im_costs;
create trigger im_cost_item_budget_finder_tr
after update or insert on im_costs
for each row execute procedure im_cost_item_budget_finder_tr();

-- update im_costs set budget_item_id = null where cost_id = 53129;
-- select cost_id, cost_name, budget_item_id from im_costs where cost_id = 53129;


update im_costs 
set budget_item_id = null
where cost_id in (
	select	c.cost_id
	from	im_costs c,
		im_projects p,
		im_projects main_p
	where	c.project_id = p.project_id and
		p.tree_sortkey between main_p.tree_sortkey and tree_right(main_p.tree_sortkey) and 
		main_p.project_id = 18228
	order by c.cost_id
--	LIMIT 1
	)
;




--  tid |	cost_type	| 
-- -----+----------------------+----------

-- Financial Docuent (classify individual invoice_items)
-- 3700 | Customer Invoice	|
-- 3702 | Quote		|
-- 3704 | Provider Bill	|
-- 3706 | Purchase Order	|
-- 3724 | Delivery Note	|
-- 3734 | Provider Receipt	|
-- 3730 | InterCo Invoice	|
-- 3732 | InterCo Quote	|

-- Default (just classify the entire thingy)
-- 3714 | Employee Salary	|
-- 3716 | Repeating Cost	|
-- 3718 | Timesheet Cost	|
-- 3726 | Timesheet Budget	|
-- 3728 | Expense Planned Cost |
-- 3736 | Timesheet Hours	|

-- Expenses speical logic
-- 3720 | Expense Item	|
-- 3722 | Expense Bundle	|


-- Should not appear at all (abstract supertype)
--
-- 3708 | Customer Documents   |
-- 3710 | Provider Documents   |

