-- /packages/intranet-budget/sql/postgresql/intranet-budget-version-create.sql
--
-- Copyright (c) 2003-2019 ]project-open[
--
-- All rights reserved. Please check
-- http://www.project-open.com/license/ for details.
--
-- @author frank.bergmann@project-open.com


-----------------------------------------------------------
-- Budget Version
--
-- Container for a number of budget objects.
-- Allows to attach an approval workflow.


SELECT acs_object_type__create_type (
	'im_budget_version',		-- object_type
	'Budget Version', 		-- pretty_name
	'Budget Versions', 		-- pretty_plural
	'acs_object',			-- supertype
	'im_budget_versions',		-- table_name
	'version_id',			-- id_column
	'intranet-budget-versions',	-- package_name
	'f',				-- abstract_p
	null,				-- type_extension_table
	'im_budget_version__name'	-- name_method
);


insert into acs_object_type_tables (object_type, table_name, id_column)
values ('im_budget_version', 'im_budget_versions', 'version_id');


-- Setup status and type columns for im_user_versions
update acs_object_types set
        status_column = 'version_status_id',
        type_column = 'version_type_id',
        status_type_table = 'im_budget_versions'
where object_type = 'im_budget_version';


insert into im_biz_object_urls (object_type, url_type, url) values ('im_budget_version','view','/intranet-budget/versions/new?display_mode=display&version_id=');
insert into im_biz_object_urls (object_type, url_type, url) values ('im_budget_version','edit','/intranet-budget/versions/new?display_mode=edit&version_id=');


create table im_budget_versions (
	version_id		integer
				constraint im_budget_versions_id_pk
				primary key
				constraint im_budget_versions_id_fk
				references acs_objects,
	version_name		text
				constraint im_budget_version_name_nn
				not null,
	version_nr		text
				constraint im_budget_version_nr_nn
				not null,
	version_object_id	integer
				constraint im_budget_version_object_fk
				references acs_objects,
	version_status_id	integer 
				constraint im_budget_versions_status_nn
				not null
				constraint im_budget_versions_status_fk
				references im_categories,
	version_type_id		integer 
				constraint im_budget_versions_type_nn
				not null
				constraint im_budget_versions_type_fk
				references im_categories,
	comment			text
);


-- Do not allow duplicate entries
create unique index im_budget_version_name_un_idx on im_budget_versions(coalesce(version_object_id,0), version_name);
create unique index im_budget_version_nr_un_idx on im_budget_versions(coalesce(version_object_id,0), version_nr);




-----------------------------------------------------------
-- Add version to budget items to keep status
--
alter table im_budget_items add column budget_item_version_id integer 
constraint im_budget_items_version_obj_fk references im_budget_versions;

-- And add an index, as access to version_obj is quite frequent.
create index im_budget_items_version_obj_idx on im_budget_items(budget_item_version_id);



-----------------------------------------------------------
-- Add version to budget items to keep status
--
alter table im_planning_items add column item_version_id integer 
constraint im_planning_items_version_obj_fk references im_budget_versions;

-- And add an index, as access to version_obj is quite frequent.
create index im_planning_items_version_obj_idx on im_planning_items(item_version_id);


-----------------------------------------------------------
-- Add version to planning_items unique constraint
--
drop index if exists im_planning_object_item_un;

create unique index im_planning_object_item_un on im_planning_items(
	item_object_id,
	coalesce(item_project_phase_id,0), 
	coalesce(item_project_member_id,0),
	coalesce(item_cost_type_id,0),
	coalesce(item_version_id,0),
	coalesce(item_date,'2000-01-01'::timestamptz)
);






-----------------------------------------------------------
-- Privileges
--

-- view_budget_versions_objects_all restricts possibility to see budget_version_objects of others
SELECT acs_privilege__create_privilege(
	'view_budget_versions_all',
	'View Budget Versions All',
	'View Budget Versions All'
);
SELECT acs_privilege__add_child('admin', 'view_budget_versions_all');


-----------------------------------------------------------
-- Create, Drop and Name Plpg/SQL functions
--
-- These functions represent crator/destructor
-- functions for the OpenACS object system.


create or replace function im_budget_version__name(integer)
returns varchar as $body$
DECLARE
	p_version_id		alias for $1;
	v_name			varchar;
BEGIN
	select	bv.version_name
	into	v_name
	from	im_budget_versions bv
	where	version_id = p_version_id;

	return v_name;
end;$body$ language 'plpgsql';


create or replace function im_budget_version__new (
	integer, varchar, timestamptz, integer, varchar, integer,
	varchar, varchar, integer, integer, integer
) returns integer as $body$
DECLARE
	p_version_id		alias for $1;
	p_object_type   	alias for $2;
	p_creation_date 	alias for $3;
	p_creation_user		alias for $4;
	p_creation_ip		alias for $5;
	p_context_id		alias for $6;

	p_version_name		alias for $7;
	p_version_nr		alias for $8;
	p_version_type_id	alias for $10;		
	p_version_status_id	alias for $11;

	v_version_id	integer;
BEGIN
	v_version_id := acs_object__new (
		p_version_id,		-- object_id
		p_object_type,		-- object_type
		p_creation_date,	-- creation_date
		p_creation_user,	-- creation_user
		p_creation_ip,		-- creation_ip
		p_context_id,		-- context_id
		't'			-- security_inherit_p
	);

	insert into im_budget_versions (
		version_id,
		version_name, version_nr,
		version_type_id, version_status_id
	) values (
		v_version_id,
		p_version_name, p_version_nr,
		p_version_type_id, p_version_status_id
	);

	return v_version_id;
END;$body$ language 'plpgsql';


create or replace function im_budget_version__delete(integer)
returns integer as $body$
DECLARE
	p_version_id	alias for $1;
BEGIN
	-- remove pointers to this object from im_budget_items
	update	im_budget_items
	set	budget_item_version_id = null
	where	budget_item_version_id = p_version_id;

	-- Delete budget tokens for cases around this object
	delete from wf_tokens 
	where case_id in (select case_id from wf_cases where object_id = p_version_id);

	-- Delete budget cases of this object.
	delete from wf_cases where object_id = p_version_id;

	-- Delete any data related to the object
	delete	from im_budget_versions
	where	version_id = p_version_id;

	-- Finally delete the object iself
	PERFORM acs_object__delete(p_version_id);

	return 0;
end;$body$ language 'plpgsql';


-----------------------------------------------------------
-- Type and Status
--
-- Create categories for Notes type and status.
-- Status acutally is not use, so we just define "active"

-- 49300-49399  Intranet Budget Version Status (100)
-- 49400-49499  Intranet Budget Version Type (100)



-- ---------------------------------------------------------
-- Budget Version Status
--
-- 49300-49399  Intranet Budget Version Status (100)
SELECT im_category_new(49300, 'Approved', 'Intranet Budget Version Status');
SELECT im_category_new(49310, 'Requested', 'Intranet Budget Version Status');
SELECT im_category_new(49320, 'Active', 'Intranet Budget Version Status');
SELECT im_category_new(49390, 'Closed', 'Intranet Budget Version Status');
SELECT im_category_new(49399, 'Deleted', 'Intranet Budget Version Status');
SELECT im_category_hierarchy_new(49399, 49390);

create or replace view im_budget_version_status as
select category_id as version_status_id, category as version_status
from im_categories
where category_type = 'Intranet Budget Version Status';



-- ---------------------------------------------------------
-- Budget Version Type
--
-- 49400-49499  Intranet Budget Version Type (100)
SELECT im_category_new(49400, 'Default', 'Intranet Budget Version Type');

create or replace view im_budget_version_type as
select category_id as version_type_id, category as version_type
from im_categories
where category_type = 'Intranet Budget Version Type';



-- ------------------------------------------------------
-- Budget Graph & Journal on Absence View Page
-- ------------------------------------------------------

SELECT  im_component_plugin__new (
	null, 'im_component_plugin', now(), null, null, null,
	'Budget Version Budget',		-- component_name
	'intranet-budget',			-- package_name
	'right',				-- location
	'/intranet-budget/versions/new',	-- page_url
	null,					-- view_name
	10,					-- sort_order
	'im_budget_graph_component -object_id $version_id'
);

SELECT  im_component_plugin__new (
	null, 'im_component_plugin', now(), null, null, null,
	'Budget Version Journal',		-- component_name
	'intranet-budget',			-- package_name
	'bottom',				-- location
	'/intranet-budget/versions/new',	-- page_url
	null,					-- view_name
	100,					-- sort_order
	'im_budget_journal_component -object_id $version_id'
);

