# /packages/intranet-budget/tcl/intranet-budget-procs.tcl
#
# Copyright (C) 2003-2007 ]project-open[
#
# All rights reserved. Please check
# http://www.project-open.com/license/ for details.

ad_library {
    @author frank.bergmann@project-open.com
}


# ----------------------------------------------------------------------
# Constants
# ----------------------------------------------------------------------

# 49000-49999  Intranet Budget (1000)


# Budget Item Status
# 49000-49099  Intranet Budget Item Status (100)
ad_proc -public im_budget_item_status_approved {} { return 49000 }
ad_proc -public im_budget_item_status_requested {} { return 49010 }
ad_proc -public im_budget_item_status_active {} { return 49020 }
ad_proc -public im_budget_item_status_closed {} { return 49090 }
ad_proc -public im_budget_item_status_deleted {} { return 49099 }

# Budget Item Types
# 49100-49199  Intranet Budget Item Type (100)
ad_proc -public im_budget_item_type_default {} { return 49100 }


# ----------------------------------------------------------------------
# PackageID
# ----------------------------------------------------------------------

ad_proc -public im_package_budget_id {} {
    Returns the package id of the intranet-budget module
} {
    return [util_memoize im_package_budget_id_helper]
}

ad_proc -private im_package_budget_id_helper {} {
    return [db_string im_package_core_id {
	select package_id from apm_packages
	where package_key = 'intranet-budget'
    } -default 0]
}


namespace eval im_budget_item {

    ad_proc -public new {
	{ -var_hash "" }
    } {
	Create a new budget item.
	@author frank.bergmann@project-open.com
	@return The object_id of the new (or existing) Budget Item.
    } {
	# Get the current user and its address.
	# If we are not connected then use some system defaults.
	if {[ns_conn isconnected]} { 
	    set peeraddr [ad_conn peeraddr] 
	    set current_user_id [ad_conn user_id]
	} else {
	    set peeraddr "0.0.0.0"
	    set current_user_id [util_memoize [list db_string first_user "select min(person_id) from persons where person_id > 0" -default 0]]
	}

	array set vars $var_hash
	set budget_item_new_sql "
		select im_budget_item__new(
			null, 'im_budget_item', now(), :current_user_id, :peeraddr, null,
			:budget_item_name,
			:budget_item_nr,
			:budget_item_parent_id,
			:budget_item_type_id,
			:budget_item_status_id,
			:budget_item_project_id
		)
	"

	# Set defaults.
	set budget_item_name $vars(budget_item_name)
	set budget_item_nr $budget_item_name
	set budget_item_parent_id ""
	set budget_item_status_id [im_budget_item_status_approved]
	set budget_item_type_id [im_budget_item_type_default]
	set budget_item_owner_id $current_user_id
	set budget_item_project_id ""
	set description ""
	set note ""

	# Override defaults
	if {[info exists vars(budget_item_nr)]} { set budget_item_nr $vars(budget_item_nr) }
	if {[info exists vars(budget_item_parent_id)]} { set budget_item_parent_id $vars(budget_item_parent_id) }
	if {[info exists vars(budget_item_status_id)]} { set budget_item_status_id $vars(budget_item_status_id) }
	if {[info exists vars(budget_item_type_id)]} { set budget_item_type_id $vars(budget_item_type_id) }
	if {[info exists vars(budget_item_owner_id)]} { set budget_item_owner_id $vars(budget_item_owner_id) }
	if {[info exists vars(budget_item_project_id)]} { set budget_item_project_id $vars(budget_item_project_id) }
	if {[info exists vars(description)]} { set description $vars(description) }
	if {[info exists vars(note)]} { set note $vars(note) }

	# Check if the item already exists
	set budget_item_id [db_string exists "
		select	budget_item_id
		from	im_budget_items
		where	budget_item_parent_id = :budget_item_parent_id and
			budget_item_nr = :budget_item_nr
	" -default 0]

	# Create a new item if necessary
	if {!$budget_item_id} { set budget_item_id [db_string new $budget_item_new_sql] }

	# Update the item with additional variables from the vars array
	set sql_list [list]
	foreach var [array names vars] {
	    if {$var eq "budget_item_id"} { continue }
	    lappend sql_list "$var = :$var"
	}
	set sql "
		update im_budget_items set
		[join $sql_list ",\n"]
		where budget_item_id = :budget_item_id
	"
	db_dml update_budget_item $sql
	return $budget_item_id
    }


    ad_proc -public audit {
	-budget_item_id:required
	-action:required
    } {
	Write the audit trail
    } {
	# Write Audit Trail
	im_audit -object_id $budget_item_id -action $action
    }


    ad_proc -public nuke {
	-budget_item_id:required
    } {
	Permanently deletes the BudgetItem from the database.
	This is only suitable for test purposes. During production operations,
	please set the BudgetItem's status to "deleted".
    } {
	return [im_budget_item_nuke -budget_item_id $budget_item_id]
    }


    ad_proc -public check_permissions {
	{-check_only_p 0}
	-budget_item_id:required
	-operation:required
    } {
	Check if the user can perform view, read, write or admin the budget_item
    } {
	set user_id [ad_conn user_id]
	set user_name [im_name_from_user_id $user_id]
	im_budget_item_permissions $user_id $budget_item_id view read write admin
	if {[lsearch {view read write admin} $operation] < 0} { 
	    ad_return_complaint 1 "Invalid operation '$operation':<br>Expected view, read, write or admin"
	    ad_script_abort
	}
	set perm [set $operation]

	# Just return the result check_only_p is set
	if {$check_only_p} { return $perm }

 	if {!$perm} { 
	    set action_forbidden_msg [lang::message::lookup "" intranet-helpdesk.Forbidden_operation_on_budget_item "
	    <b>Unable to perform operation '%operation%'</b>:<br>You don't have the necessary permissions for budget_item #%budget_item_id%."]
	    ad_return_complaint 1 $action_forbidden_msg 
	    ad_script_abort
	}
	return $perm
    }
}




# ----------------------------------------------------------------------
# Generate generic select SQL for Budget Items
# to be used in list pages, options, ...
# ---------------------------------------------------------------------

ad_proc -public im_budget_item_select_sql { 
    {-type_id ""} 
    {-status_id ""} 
    {-object_id ""} 
    {-owner_id ""} 
    {-member_id ""}
    {-parent_id ""}
    {-var_list "" }
    {-treelevel ""}
} {
    Returns an SQL statement that allows you to select a range of
    budget items, given a number of conditions.
    This SQL is used for example in the BudgetItemListPage, in
    im_budget_item_options and others.
    The variable names returned by the SQL adhere to the ]po[ coding
    standards. Important returned variables include:
	- im_budget_items.*, (all fields from the base table)
	- budget_item_status, budget_item_type, (status and type human readable)
    The result set is a conjunction (and-connection) of the specified
    conditions:
	- type_id & status_id: Limit type and status, including sub-types and states
	- object_id: Returns CIs associated with project
	- owner_id: Checks for the specific budget_item_owner_id
	- member_id: Check for owner_id OR association relationship.
} {
    # Prepare and check some variables.
    set current_user_id 0
    if {[ns_conn isconnected]} { set current_user_id [ad_conn user_id] }

    # base url, where only the budget_item_id has to be added
    set budget_item_base_url "/intranet-budget/new?form_mode=display&budget_item_id="

    im_security_alert_check_integer -location im_budget_item_select_sql -value $type_id
    im_security_alert_check_integer -location im_budget_item_select_sql -value $status_id
    im_security_alert_check_integer -location im_budget_item_select_sql -value $object_id
    im_security_alert_check_integer -location im_budget_item_select_sql -value $member_id
    im_security_alert_check_integer -location im_budget_item_select_sql -value $owner_id
    im_security_alert_check_integer -location im_budget_item_select_sql -value $parent_id
    im_security_alert_check_integer -location im_budget_item_select_sql -value $treelevel

    # Deal with generically passed variables as replacement of parameters.
    array set var_hash $var_list
    foreach var_name [array names var_hash] { set $var_name $var_hash($var_name) }

    set extra_froms [list]
    set extra_wheres [list]

    # Member: Either owner or membership_rel
    if {"" != $member_id && 0 != $member_id} {
	lappend extra_wheres "(
		i.budget_item_owner_id = $member_id
	OR	i.budget_item_id in (
		select	r.object_id_two
		from	acs_rels r
		where	r.object_id_one = $member_id
	))"
    }

    # Owner: Check for budget_item_owner_id field
    if {"" != $owner_id && 0 != $owner_id} {
	lappend extra_wheres "i.budget_item_owner_id = $owner_id"
    }

    if {"" != $object_id && 0 != $object_id} {
	lappend extra_wheres "i.budget_item_id in (
		-- BIs related to the current project directly
		select	sub_bi.budget_item_id
		from	im_budget_items bi,
			im_budget_items sub_bi
		where	bi.budget_item_project_id = $object_id and
			sub_bi.tree_sortkey between bi.tree_sortkey and tree_right(bi.tree_sortkey)
UNION
		-- BIs related to the current project
		select	sub_bi.budget_item_id
		from	acs_rels bir,
			im_budget_item_project_rels bipr,
			im_budget_items bi,
			im_budget_items sub_bi
		where	bir.object_id_one = $object_id and
			bir.object_id_two = bi.budget_item_id and
			bir.rel_id = bipr.rel_id and
			sub_bi.tree_sortkey between bi.tree_sortkey and tree_right(bi.tree_sortkey)

        )"
    }

    # -----------------------------------------------
    # Permissions

    set perm_where "
	('t' = acs_permission__permission_p([subsite::main_site_id], $current_user_id, 'view_budget_items_all') OR
	i.budget_item_id in (
		-- User is explicit member of budget item
		select	bi.budget_item_id
		from	im_budget_items bi,
			acs_rels r
		where	r.object_id_two = $current_user_id and
			r.object_id_one = bi.budget_item_id
	UNION
		-- User is a member of a group that is explicit member of budget item
		select	bi.budget_item_id
		from	im_budget_items bi,
			acs_rels r
		where	r.object_id_one = bi.budget_item_id and
			r.object_id_two in (
				select	group_id
				from	group_distinct_member_map
				where	member_id = $current_user_id
			)
	UNION
		-- User belongs to project that belongs to budget item
		select	bi.budget_item_id
		from	im_budget_items bi,
			im_projects p,
			acs_rels r1,
			acs_rels r2
		where	r1.object_id_two = $current_user_id and
			r1.object_id_one = p.project_id and
			r2.object_id_two = bi.budget_item_id and
			r2.object_id_one = p.project_id
	UNION
		-- User belongs to project that belongs to budget item
		select	sub_bi.budget_item_id
		from	acs_rels r1,
			im_projects p,
			im_projects sub_p,
			acs_rels r2,
			im_budget_items bi,
			im_budget_items sub_bi
		where	r1.object_id_two = $current_user_id and
			r1.object_id_one = p.project_id and
			sub_p.tree_sortkey between p.tree_sortkey and tree_right(p.tree_sortkey) and
			r2.object_id_one = sub_p.project_id and
			r2.object_id_two = bi.budget_item_id and
			sub_bi.tree_sortkey between bi.tree_sortkey and tree_right(bi.tree_sortkey)
	))
    "

    set project_perm_sql "
	select  pp.project_id,
		bom.object_role_id as role_id
	from
		im_projects pp,
		acs_rels r,
		im_biz_object_members bom
	where
		r.object_id_one = pp.project_id and
		r.object_id_two = :user_id and
		r.rel_id = bom.rel_id and
		pp.tree_sortkey in (
			-- Walk up the budget item is-part-of hierarchy and collect
			-- all projects for which the BI is a member.
			-- Returns the list of all these projects and their parents.
			select  tree_ancestor_keys(p.tree_sortkey)
			from    im_budget_items bi,
				acs_rels r,
				im_projects p
			where   r.object_id_two = bi.budget_item_id and
				r.object_id_one = p.project_id and
				bi.tree_sortkey in (
					select  tree_ancestor_keys(sbi.tree_sortkey)
					from    im_budget_items sbi
					where   sbi.budget_item_id = :budget_item_id
				)
			)
    "


    # -----------------------------------------------
    # Join the query parts

    if {"" != $status_id} { lappend extra_wheres "i.budget_item_status_id in ([join [im_sub_categories $status_id] ","])" }
    if {"" != $type_id} { lappend extra_wheres "i.budget_item_type_id in ([join [im_sub_categories $type_id] ","])" }
    if {"" != $treelevel} { lappend extra_wheres "tree_level(i.tree_sortkey) <= 1+$treelevel" }
    if {"" != $perm_where} { lappend extra_wheres $perm_where }
    if {"" != $parent_id} { 
	lappend extra_wheres "parent.budget_item_id = $parent_id" 
	lappend extra_wheres "i.tree_sortkey between parent.tree_sortkey and tree_right(parent.tree_sortkey)" 
	lappend extra_wheres "i.budget_item_id != parent.budget_item_id" 
	lappend extra_froms "im_budget_items parent"
    }

    set extra_from [join $extra_froms "\n\t\t,"]
    set extra_where [join $extra_wheres "\n\t\tand "]

    if {"" != $extra_from} { set extra_from ",$extra_from" }
    if {"" != $extra_where} { set extra_where "and $extra_where" }

    set select_sql "
	select distinct
		i.*,
		tree_level(i.tree_sortkey)-1 as budget_item_level,
		im_category_from_id(i.budget_item_status_id) as budget_item_status,
		im_category_from_id(i.budget_item_type_id) as budget_item_type,
		im_budget_item_name_from_id(i.budget_item_parent_id) as budget_item_parent,
		im_name_from_user_id(i.budget_item_owner_id) as budget_item_owner,
		acs_object__name(i.budget_item_project_id) as budget_item_project_name,
		'$budget_item_base_url' || i.budget_item_id as budget_item_url
	from	im_budget_items i	
		$extra_from
	where	1=1 
		$extra_where
	order by
		i.tree_sortkey
    "

    return $select_sql
}

ad_proc -public im_budget_item_update_sql { 
    {-include_dynfields_p 0}
} {
    Returns an SQL statement that updates all Budget Item fields from
    variables according to the ]po[ coding conventions.
} {
    set update_sql "
	update im_budget_items set
		budget_item_name =		:budget_item_name,
		budget_item_nr =		:budget_item_nr,
		budget_item_parent_id =		:budget_item_parent_id,
		budget_item_type_id =		:budget_item_type_id,
		budget_item_status_id =		:budget_item_status_id,
		budget_item_owner_id =		:budget_item_owner_id,
		description = 			:description,
		note = 				:note
	where budget_item_id = :budget_item_id
    "
}


# ----------------------------------------------------------------------
# Delete a budget item
# ---------------------------------------------------------------------


# ----------------------------------------------------------------------
# Nuke a Budget Item
# ---------------------------------------------------------------------

ad_proc -public im_budget_item_nuke {
    -budget_item_id:required
} {
    Permanently deletes the BudgetItem from the database.
    This is only suitable for test purposes. During production operations,
    please set the BudgetItem's status to "deleted".
} {
    set parent_p [db_string parent "select count(*) from im_budget_items where budget_item_parent_id = :budget_item_id"]
    if {$parent_p > 0} { ad_return_complaint 1 "<b>Can't Delete Budget Item</b>:<br>The budget item is the parent of another budget item. <br>Please delete the children first." }

    # Relationships
    ns_log Notice "im_budget_item_nuke: rels"
    set rels [db_list rels "
		select rel_id 
		from acs_rels 
		where object_id_one = :budget_item_id 
			or object_id_two = :budget_item_id
    "]

    set im_budget_item_project_rels_exists_p [im_table_exists im_budget_item_project_rels]
    foreach rel_id $rels {
	db_dml del_rels "delete from group_element_index where rel_id = :rel_id"
	if {[im_column_exists im_biz_object_members skill_profile_rel_id]} {
	    db_dml del_rels "update im_biz_object_members set skill_profile_rel_id = null where skill_profile_rel_id = :rel_id"
	}
	if {[im_table_exists im_gantt_assignment_timephases]} {
	    db_dml del_rels "delete from im_gantt_assignment_timephases where rel_id = :rel_id"
	}
	if {[im_table_exists im_gantt_assignments]} {
	    db_dml del_rels "delete from im_gantt_assignments where rel_id = :rel_id"
	}
	if {[im_table_exists im_agile_task_rels]} {
	    db_dml del_rels "delete from im_agile_task_rels where rel_id = :rel_id"
	}
	db_dml del_rels "delete from im_biz_object_members where rel_id = :rel_id"
	db_dml del_rels "delete from membership_rels where rel_id = :rel_id"
	if {$im_budget_item_project_rels_exists_p} { db_dml del_rels "delete from im_budget_item_project_rels where rel_id = :rel_id" }
	if {[im_table_exists im_release_items]} {
	    db_dml del_rels "delete from im_release_items where rel_id = :rel_id"
	}
	db_dml del_rels "delete from acs_rels where rel_id = :rel_id"
	db_dml del_rels "delete from acs_objects where object_id = :rel_id"
    }

    set conf_proj_rels_sql "
		select	r.*
		from	acs_rels r,
			im_projects p
		where	object_id_one = p.project_id
			and object_id_two = :budget_item_id
    "
    db_foreach conf_proj_rels_del $conf_proj_rels_sql {
	db_string del_conf_proj_rel "select im_budget_item_project_rel__delete(:rel_id)"
    }
    
    db_dml nuke_bi_context_id "update acs_objects set context_id = null where context_id = :budget_item_id"
    db_string nuke_bi "select im_budget_item__delete(:budget_item_id)"
}

# ----------------------------------------------------------------------
# Permissions
# ---------------------------------------------------------------------


ad_proc -public im_budget_item_permissions {user_id budget_item_id view_var read_var write_var admin_var} {
    Fill the "by-reference" variables read, write and admin
    with the permissions of $user_id on $budget_item_id
} {
    upvar $view_var view
    upvar $read_var read
    upvar $write_var write
    upvar $admin_var admin

    set read [im_permission $user_id view_budget_items_all]
    set write [im_permission $user_id edit_budget_items_all]
    set admin [im_permission $user_id edit_budget_items_all]

    set user_is_admin_p [im_is_user_site_wide_or_intranet_admin $user_id]
    set user_is_wheel_p [im_profile::member_p -profile_id [im_wheel_group_id] -user_id $user_id]
    set user_is_group_member_p [im_biz_object_member_p $user_id $budget_item_id]
    set user_is_group_admin_p [im_biz_object_admin_p $user_id $budget_item_id]

    # Admin permissions to global + intranet admins + group administrators
    set user_admin_p [expr {$user_is_admin_p || $user_is_group_admin_p || $user_is_wheel_p}]
    if {$user_admin_p} {
	set read 1
	set write 1
	set admin 1
    }

    if {$user_is_group_member_p} {
	set read 1
    }

    # Project-based permissions - Check if the current user is an explicit member
    # of a project, to which the BudgetItem belongs
    # Normal membership of the project is sufficient to get write permission on the budget item(?)
    set project_perm_sql "
	select  pp.project_id,
		bom.object_role_id as role_id
	from
		im_projects pp,
		acs_rels r,
		im_biz_object_members bom
	where
		r.object_id_one = pp.project_id and
		r.object_id_two = :user_id and
		r.rel_id = bom.rel_id and
		pp.tree_sortkey in (
			-- Walk up the budget item is-part-of hierarchy and collect
			-- all projects for which the BI is a member.
			select  tree_ancestor_keys(p.tree_sortkey)
			from    im_budget_items bi,
				im_projects p
			where   bi.budget_item_project_id = p.project_id and
				bi.tree_sortkey in (
					select  tree_ancestor_keys(sbi.tree_sortkey)
					from    im_budget_items sbi
					where   sbi.budget_item_id = :budget_item_id
				)
			)
    "
    # ad_return_complaint 1 "<pre>[im_ad_hoc_query $project_perm_sql]</pre>"
    db_foreach project_perms $project_perm_sql {
	# normal project members only get read permissions on the budget items
	set read 1
	# PMs get write permissions on the budget items
	if {[im_biz_object_role_project_manager] eq $role_id} { 
	    set write 1 
	}
    }

    # Tricky: Check if the user is the owner of one of the parent BIs...
    # ToDo: not yet implemented...

    # No explict view perms - set to tread
    set view $read

    # No read - no write...
    if {!$read} {
	set write 0
	set admin 0
    }

    # ad_return_complaint 1 "$read $write $admin"
}


# ----------------------------------------------------------------------
# Configurable list of Budget Items
# ---------------------------------------------------------------------

ad_proc -public im_budget_item_list_component {
    {-debug 0}
    {-object_id 0}
    {-owner_id 0}
    {-member_id 0}
    {-view_name "im_budget_item_list"} 
    {-order_by ""} 
    {-restrict_to_member_id 0} 
    {-restrict_to_type_id 0} 
    {-restrict_to_status_id 0} 
    {-max_entries_per_page 5000}
    {-export_var_list {} }
    {-return_url "" }
} {
    Creates a HTML table showing a list of budget items associated with
    a project or another object.
} {
    # ---------------------- Security - Show the comp? -------------------------------
    set current_user_id [auth::require_login]
    set user_is_admin_p [im_is_user_site_wide_or_intranet_admin $current_user_id]
    set org_object_id $object_id

    if {"" == $member_id || 0 == $member_id} { set member_id $restrict_to_member_id }
    if {"" == $order_by} {
	set order_by [parameter::get_from_package_key -package_key intranet-budget -parameter BudgetItemComponentDefaultSortOrder -default "tree_sortkey"] 
    }

    # URL to toggle open/closed tree
    set open_close_url "/intranet/biz-object-tree-open-close"    

    # Permissions
    set object_view 0
    set object_read 0
    set object_write 0
    set object_admin 0
    set object_type [db_string acs_object_type "select object_type from acs_objects where object_id = :object_id" -default ""]
    if {"" != $object_type} {
	set perm_cmd "${object_type}_permissions \$current_user_id \$object_id object_view object_read object_write object_admin"
	eval $perm_cmd
    }


    # ---------------------- Defaults ----------------------------------

    # Get the start_idx in case of pagination
    set start_idx [im_opt_val -limit_to integer "budget_item_start_idx"]
    if {"" == $start_idx} { set start_idx 0 }
    set end_idx [expr {$start_idx + $max_entries_per_page - 1}]

    set bgcolor(0) " class=roweven"
    set bgcolor(1) " class=rowodd"
    set date_format "YYYY-MM-DD"

    set current_url [im_url_with_query]

    if {![info exists current_page_url]} { set current_page_url [ad_conn url] }
    if {(![info exists return_url] || $return_url eq "")} { set return_url $current_url }
    # Get the "view" (=list of columns to show)
    set view_id [db_string get_view_id "select view_id from im_views where view_name = :view_name" -default 0]

    if {0 == $view_id} {
	ns_log Error "im_budget_item_list_component: we didn't find view_name=$view_name"
	set view_name "im_budget_item_list"
	set view_id [db_string get_view_id "select view_id from im_views where view_name=:view_name" -default 0]
    }

    if {$debug} { ns_log Notice "im_budget_item_list_component: view_id=$view_id" }
    if {0 == $view_id} {
	return [lang::message::lookup "" intranet-core.lt_Internal_error_unknow "Unknown View: $view_name, please inform your System Adminitrator"]
    }

    # ---------------------- Get Columns ----------------------------------
    # Define the column headers and column contents that
    # we want to show:
    #
    set column_headers [list]
    set column_vars [list]
    set admin_links [list]
    set extra_selects [list]
    set extra_froms [list]
    set extra_wheres [list]
    set view_order_by_clause ""

    set column_sql "
	select	*
	from	im_view_columns
	where	view_id = :view_id
		and group_id is null
	order by sort_order
    "
    set col_span 0
    db_foreach column_list_sql $column_sql {
	if {"" == $visible_for || [eval $visible_for]} {
	    lappend column_headers "$column_name"
	    lappend column_vars "$column_render_tcl"
	    lappend admin_links "<a href=[export_vars -base "/intranet/admin/views/new-column" {return_url column_id {form_mode edit}}] target=\"_blank\">[im_gif wrench]</a>"

	    if {"" != $extra_select} { lappend extra_selects $extra_select }
	    if {"" != $extra_from} { lappend extra_froms $extra_from }
	    if {"" != $extra_where} { lappend extra_wheres $extra_where }
	    if {"" != $order_by_clause && $order_by == $column_name} { set view_order_by_clause $order_by_clause }
	}
	incr col_span
    }
    if {$debug} { ns_log Notice "im_budget_item_list_component: column_headers=$column_headers" }

    # -------- Compile the list of parameters to pass-through-------
    set bind_vars [ns_set create]
    foreach var $export_var_list {
	set value [im_opt_val -limit_to nohtml $var]
	if {$value ne ""} {
	    ns_set put $bind_vars $var $value
	    if {$debug} { ns_log Notice "im_budget_item_list_component: $var <- $value" }
	}
    }

    ns_set delkey $bind_vars "order_by"
    ns_set delkey $bind_vars "budget_item_start_idx"
    set params [list]
    set len [ns_set size $bind_vars]
    for {set i 0} {$i < $len} {incr i} {
	set key [ns_set key $bind_vars $i]
	set value [ns_set value $bind_vars $i]
	if {$value ne "" } {
	    lappend params "$key=[ns_urlencode $value]"
	}
    }
    set pass_through_vars_html [join $params "&"]


    # ---------------------- Format Header ----------------------------------
    # Set up colspan to be the number of headers + 1 for the # column
    set colspan [expr {[llength $column_headers] + 1}]

    # Format the header names with links that modify the
    # sort order of the SQL query.
    #
    set col_ctr 0
    set admin_link ""
    set table_header_html ""
    foreach col $column_headers {
	set cmd_eval ""
	set cmd "set cmd_eval $col"
	ns_log Notice "im_budget_item_list_component: cmd='$cmd'"
	eval $cmd
	regsub -all " " $cmd_eval "_" cmd_eval_subs

	# Only localize "reasonable" strings...
	if {[regexp {^[a-zA-Z0-9_\.\ ]+$} $cmd_eval_subs]} {
	    set cmd_eval [lang::message::lookup "" intranet-timesheet2-tasks.$cmd_eval_subs $cmd_eval]
	}

	if {$user_is_admin_p} { set admin_link [lindex $admin_links $col_ctr] } else { set admin_link "" }
	append table_header_html "  <td class=\"rowtitle\">$cmd_eval$admin_link</td>\n"
	incr col_ctr
    }

    set table_header_html "
	<thead>
	    <tr class=tableheader>$table_header_html</tr>
	</thead>
    "
    
    # ---------------------- Build the SQL query ---------------------------
    set order_by_clause "order by bi.tree_sortkey"
    set order_by_clause_ext "order by budget_item_nr, budget_item_name"
    switch $order_by {
	"Status" { 
	    set order_by_clause "order by bi.budget_item_status_id" 
	    set order_by_clause_ext "m.budget_item_id"
	}
    }
	
    # ---------------------- Calculate the Children's restrictions -------------------------
    set criteria [list]

    if {[string is integer $restrict_to_status_id] && $restrict_to_status_id > 0} {
	lappend criteria "sub_bi.budget_item_status_id in ([join [im_sub_categories $restrict_to_status_id] ","])"
    }

    if {[string is integer $restrict_to_type_id] && $restrict_to_type_id > 0} {
	lappend criteria "sub_bi.budget_item_type_id in ([join [im_sub_categories $restrict_to_type_id] ","])"
    }

    # Owner is stictly the owner_id of the budget_item
    if {[string is integer $owner_id] && $owner_id > 0} {
	lappend criteria "sub_bi.budget_item_owner_id = :owner_id"
    }

    # Member is anybody associated with the budget item + the owner.
    # Challenge: We also need to show the parents of sub-items with that member.
    # We assume that there are few BudgetItems per user only...
    if {[string is integer $member_id] && $member_id > 0} {

	# Get the list of BIs who are directly associated with $member_id
	set new_parents [db_list user_budget_items "
		select	bi.budget_item_id
		from	im_budget_items bi,
			acs_rels r
		where	r.object_id_one = bi.budget_item_id and
			r.object_id_two = :member_id
	UNION
		select	bi.budget_item_id
		from	im_budget_items bi
		where	bi.budget_item_owner_id = :member_id
	"]

	# Loop through all parents of member_id BIs
	set result_list [list 0]
	set cnt 0
	while {[llength $new_parents] > 0} {
	    set result_list [concat $result_list $new_parents]
	    set new_parents [db_list new_parents "
		select distinct
			budget_item_parent_id
		from	im_budget_items
		where	budget_item_parent_id is not null and
			budget_item_id in ([join $new_parents ","])
	    "]
	    incr cnt
	    if {$cnt > 10} { ad_return_complaint 1 "im_budget_item_list_component: Infinite loop looking for parent: '$new_parents'" }
	}

	# Restrict to this list of direct members and their parents
	lappend criteria "sub_bi.budget_item_id in ([join $result_list ","])"
    }

    if {![im_permission $current_user_id "view_budget_items_all"]} {
	lappend criteria "sub_bi.budget_item_id in (
			select	bi.budget_item_id
			from	im_budget_items bi,
				acs_rels r
			where	r.object_id_one = bi.budget_item_id and 
				r.object_id_two = :current_user_id
			)
	"
	lappend criteria "main_bi.budget_item_id in (
			select	bi.budget_item_id
			from	im_budget_items bi,
				acs_rels r
			where	r.object_id_one = bi.budget_item_id and 
				r.object_id_two = :current_user_id
			)
	"
    }

    set restriction_clause [join $criteria "\n\tand "]
    if {"" != $restriction_clause} { 
	set restriction_clause "and $restriction_clause" 
    }

    set extra_select [join $extra_selects ",\n\t"]
    if { $extra_select ne "" } { set extra_select ",\n\t$extra_select" }

    set extra_from [join $extra_froms ",\n\t"]
    if { $extra_from ne "" } { set extra_from ",\n\t$extra_from" }

    set extra_where [join $extra_wheres "and\n\t"]
    if { $extra_where ne "" } { set extra_where "and \n\t$extra_where" }


    # ---------------------- Get the SQL Query -------------------------

    set budget_item_sql [im_budget_item_select_sql \
		 -object_id $org_object_id \
		 -treelevel "" \
		 -type_id "" \
		 -status_id "" \
		 -owner_id "" \
    ]
    # ad_return_complaint 1 "<pre>$budget_item_sql</pre>"
    # ad_return_complaint 1 "[im_ad_hoc_query -format html $budget_item_sql]"

    db_multirow budget_item_list_multirow budget_item_list_sql $budget_item_sql {

	# Perform the following steps in addition to calculating the multirow:
	# The list of all budget_items
	set all_budget_items_hash($budget_item_id) 1

	# The list of budget_items that have a sub-budget_item
	set parents_hash($budget_item_parent_id) 1
    }

    # ----------------------------------------------------
    # Determine closed BIs and their children

    # Store results in hash array for faster join
    # Only store positive "closed" branches in the hash to save space+time.
    # Determine the sub-budget_items that are also closed.
    set oc_sub_sql "
	select	child.budget_item_id as child_id
	from	im_budget_items child,
		im_budget_items parent
	where	parent.budget_item_id in (
			select	ohs.object_id
			from	im_biz_object_tree_status ohs
			where	ohs.open_p = 'c' and
				ohs.user_id = :current_user_id and
				ohs.page_url = 'default' and
				ohs.object_id in (
					select	budget_item_id
					from	($budget_item_sql) t
				)
			) and
		child.tree_sortkey between parent.tree_sortkey and tree_right(parent.tree_sortkey)
    "
    db_foreach oc_sub $oc_sub_sql {
	set closed_budget_items_hash($child_id) 1
    }

    # Calculate the list of leaf budget_items
    set all_budget_items_list [array names all_budget_items_hash]
    set parents_list [array names parents_hash]
    set leafs_list [set_difference $all_budget_items_list $parents_list]
    foreach leaf_id $leafs_list { set leafs_hash($leaf_id) 1 }

    if {$debug} { 
	ns_log Notice "im_budget_item_list_component: all_budget_items_list=$all_budget_items_list"
	ns_log Notice "im_budget_item_list_component: parents_list=$parents_list"
	ns_log Notice "im_budget_item_list_component: leafs_list=$leafs_list"
	ns_log Notice "im_budget_item_list_component: closed_budget_items_list=[array get closed_budget_items_hash]"
	ns_log Notice "im_budget_item_list_component: "
    }

    # Render the multirow
    set table_body_html ""
    set ctr 0
    set idx $start_idx
    set old_budget_item_id 0

    # ----------------------------------------------------
    # Render the list of BIs
    template::multirow foreach budget_item_list_multirow {

	# Skip this entry completely if the parent of this budget_item is closed
	if {[info exists closed_budget_items_hash($budget_item_parent_id)]} { continue }

	set indent_html ""
	set indent_short_html ""
	for {set i 0} {$i < $budget_item_level} {incr i} {
	    append indent_html "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"
	    append indent_short_html "&nbsp;&nbsp;&nbsp;"
	}

	if {[info exists closed_budget_items_hash($budget_item_id)]} {
	    # Closed budget_item
	    set gif_html "<a href='[export_vars -base $open_close_url {{user_id $current_user_id} {page_url "default"} {object_id $budget_item_id} {open_p "o"} return_url}]'>[im_gif "plus_9"]</a>"
	} else {
	    # So this is an open budget_item - show a "(-)", unless the budget_item is a leaf.
	    set gif_html "<a href='[export_vars -base $open_close_url {{user_id $current_user_id} {page_url "default"} {object_id $budget_item_id} {open_p "c"} return_url}]'>[im_gif "minus_9"]</a>"
	    if {[info exists leafs_hash($budget_item_id)]} { set gif_html "&nbsp;" }
	}

	set object_url [export_vars -base "/intranet-budget/new" {{budget_item_id $budget_item_id} {form_mode display} return_url}]

	# Table fields for timesheet budget_items
	set status_select [im_category_select {Intranet Budget Item Status} budget_item_status_id.$budget_item_id $budget_item_status_id]

	set budget_item_name "<nobr>[string range $budget_item_name 0 40]</nobr>"

	# We've got a budget_item.
	# Write out a line with budget_item information
	append table_body_html "<tr$bgcolor([expr {$ctr % 2}])>\n"

	foreach column_var $column_vars {
	    append table_body_html "\t<td valign=top>"
	    set cmd "append table_body_html $column_var"
	    eval $cmd
	    append table_body_html "</td>\n"
	}
	append table_body_html "</tr>\n"

	# Update the counter.
	incr ctr
	if { $max_entries_per_page > 0 && $ctr >= $max_entries_per_page } {
	    set more_url [export_vars -base "/intranet-budget/index" {{budget_item_id $object_id} {view_name "im_budget_item_budget_item_list"}}]
	    append table_body_html "
		<tr><td colspan=99>
		<b>[lang::message::lookup "" intranet-budget.List_cut_at_n_entries "List cut at %max_entries_per_page% entries"]</b>.
		[lang::message::lookup "" intranet-budget.List_cut_at_n_entries_msg "
			Please click <a href=%more_url%>here</a> for the entire list.
		"]
		</td></tr>\n"
	    break
	}
    }

    # ----------------------------------------------------
    # Show a reasonable message when there are no result rows:
    if { $table_body_html eq "" } {
	set table_body_html "
		<tr class=table_list_page_plain>
			<td colspan=$colspan align=left>
			<b>[lang::message::lookup "" intranet-budget.There_are_no_active_budget_items "There are not active budget items"]</b>
			</td>
		</tr>
	"
    }


    if { "im_project" == [acs_object_type $org_object_id] } {
	set new_budget_item_url [export_vars -base "/intranet-budget/new" {{form_mode edit} {return_url $current_url} {budget_item_project_id $org_object_id}}]
    } else {
	set new_budget_item_url [export_vars -base "/intranet-budget/new" {{form_mode edit} {return_url $current_url}}]
    }

    set table_body_html_ul ""
    if {[im_permission $current_user_id "add_budget_items"]} {
	append table_body_html_ul "
		<li><a href=\"$new_budget_item_url\">[lang::message::lookup "" intranet-budget.Create_New_Conf_Item "Create new Budget Item"]</a>
        "
    }

    set table_action_html "
	<tr>
	<td colspan=$colspan>
	<ul>
                $table_body_html_ul
	</ul>
	</td>
	</tr>
    "
    if {"" eq $table_body_html_ul} { set table_action_html "" }
    
    set total_in_limited 0

    # Deal with pagination
    if {$ctr == $max_entries_per_page && $end_idx < [expr {$total_in_limited - 1}]} {
	# This means that there are rows that we decided not to return
	# Include a link to go to the next page
	set next_start_idx [expr {$end_idx + 1}]
	set budget_item_max_entries_per_page $max_entries_per_page
	set next_page_url  "[export_vars -base $current_page_url {budget_item_id budget_item_project_id budget_item_max_entries_per_page order_by}]&budget_item_start_idx=$next_start_idx&$pass_through_vars_html"
	set next_page_html "($remaining_items more) <A href=\"$next_page_url\">&gt;&gt;</a>"
    } else {
	set next_page_html ""
    }
    
    if { $start_idx > 0 } {
	# This means we didn't start with the first row - there is
	# at least 1 previous row. add a previous page link
	set previous_start_idx [expr {$start_idx - $max_entries_per_page}]
	if { $previous_start_idx < 0 } { set previous_start_idx 0 }
	set previous_page_html "<A href=[export_vars -base $current_page_url {budget_item_id}]&$pass_through_vars_html&order_by=$order_by&budget_item_start_idx=$previous_start_idx>&lt;&lt;</a>"
    } else {
	set previous_page_html ""
    }
    

    # ---------------------- Format the action bar at the bottom ------------

    set table_footer_action "
	<table width='100%'>
	<tr>
	<td align=left>
		[im_category_select "Intranet Budget Item Action" action_id ""]
		<input type=submit name=submit value='[lang::message::lookup "" intranet-budget.Apply Apply]'>
	</td>
	</tr>
	</table>
    "

    set table_footer "
	<tfoot>
	<tr>
	  <td class=rowplain colspan=$colspan align=right>
	    $previous_page_html
	    $next_page_html
	    $table_footer_action
	    $table_action_html
	  </td>
	</tr>
	<tfoot>
    "

    # ---------------------- Join all parts together ------------------------

    set component_html "
	<form action=/intranet-budget/action method=GET>
	[export_vars -form {{budget_item_project_id $object_id} return_url}]
	<table bgcolor=white border=0 cellpadding=1 cellspacing=1 class=\"table_list_page\">
	  $table_header_html
	  $table_body_html
	  $table_footer
	</table>
	</form>
    "

    return $component_html
}


# ----------------------------------------------------------------------
# Options for ad_form
# ---------------------------------------------------------------------

ad_proc -public im_budget_item_options { 
    {-include_empty_p 0} 
    {-include_empty_name ""} 
    {-type_id ""} 
    {-status_id ""} 
    {-project_id ""} 
    {-owner_id ""} 
    {-cost_center_id ""} 
} {
    Returns a list of all Budget Items.
} {
    set var_list [list type_id $type_id status_id $status_id project_id $project_id owner_id $owner_id cost_center_id $cost_center_id]
    set options_sql [im_budget_item_select_sql -treelevel 2 -var_list $var_list]

    set options [list]
    if {$include_empty_p} { lappend options [list $include_empty_name ""] }

    set cnt 0
    db_foreach budget_item_options $options_sql {
	set budget_item_name [string range $budget_item_name 0 69 ]
	set spaces ""
	for {set i 0} {$i < $budget_item_level} { incr i } {
	    append spaces "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"
	}
	lappend options [list "$spaces$budget_item_name" $budget_item_id]
	incr cnt
    }

    if {!$cnt && $include_empty_p} {
	set not_found [lang::message::lookup "" intranet-budget.No_Conf_Items_Found "No Budget Items found"]
	lappend options [list $not_found ""]
    }

    return $options
}



# ---------------------------------------------------------------
# Component showing related objects
# ---------------------------------------------------------------

ad_proc -public im_budget_item_related_objects_component {
    -budget_item_id:required
} {
    Returns a HTML component with the list of related objects.
} {
    set params [list \
		    [list base_url "/intranet-budget/"] \
		    [list budget_item_id $budget_item_id] \
		    [list return_url [im_url_with_query]] \
    ]

    set result [ad_parse_template -params $params "/packages/intranet-budget/lib/related-objects-component"]
    return [string trim $result]
}





ad_proc -public im_budget_tree_component {
    -project_id:required
    {-diagram_width 800 }
    {-diagram_height 400 }
    {-diagram_title "" }
} {
    Returns a HTML component with a Sencha tree of budget items
} {
    if {[im_security_alert_check_integer -location "im_ganttproject_gantt_component" -value $project_id]} { return "" }
    set project_type_id [util_memoize [list db_string project_type "select project_type_id from im_projects where project_id = $project_id" -default ""]]

    # Only show with main projects
    set parent_id [db_string parent "select parent_id from im_projects where project_id = :project_id" -default 0]
    if {"" ne $parent_id} { return "" }

    # Sencha check and permissions
    if {![im_sencha_extjs_installed_p]} { return "" }
    im_sencha_extjs_load_libraries

    set params [list \
		    [list project_id $project_id] \
		    [list diagram_width $diagram_width] \
		    [list diagram_height $diagram_height] \
		    [list diagram_title $diagram_title] \
    ]

    set result [ad_parse_template -params $params "/packages/intranet-budget/lib/budget-tree-component"]
    return [string trim $result]
}





# ----------------------------------------------------------------------
# Budget Item - Project Relationship
# ---------------------------------------------------------------------

ad_proc -public im_budget_item_new_project_rel {
    -project_id:required
    -budget_item_id:required
    {-sort_order 0}
} {
    Establishes as is-budget-item-of relation between budget item and project
} {
    if {"" == $project_id} { ad_return_complaint 1 "Internal Error - project_id is NULL" }
    if {"" == $budget_item_id} { ad_return_complaint 1 "Internal Error - budget_item_id is NULL" }

    set rel_id [db_string rel_exists "
	select	rel_id
	from	acs_rels
	where	object_id_one = :project_id
		and object_id_two = :budget_item_id
    " -default 0]
    if {0 != $rel_id} { return $rel_id }

    return [db_string new-budget-project_rel "
	select im_budget_item_project_rel__new (
		null,
		'im_budget_item_project_rel',
		:project_id,
		:budget_item_id,
		null,
		[ad_conn user_id],
		'[ad_conn peeraddr]',
		:sort_order
	)
    "]
}




# ---------------------------------------------------------------
# 
# ---------------------------------------------------------------

ad_proc -public im_budget_item_rollup_cost_type_ids {
} {
    Returns a list of cost types that are suitable to be rolled-up
    from individual cost items up to budget items
} {
    # Excluding:
    # - Customer documents (3700, 3702)
    # - Budget types (3726, 3728)
    # - Abstract types (3708, 3710)
    # - Obsolete types (3714, 3716)
    # - InterCo types (3730, 3732)
    # - Employee Salary (3714)
    # - Delivery Note (3724)
    set rollup_cost_type_ids {3706 3734 3704 3718 3720 3722}
    return $rollup_cost_type_ids
}


