# /packages/intranet-budget/www/data-source/budget-tree-action.tcl
#
# Copyright (C) 2013-2020 ]project-open[


# ---------------------------------------------------------------
# XHR POST - ad_page_contract doesn't parse URL vars, apparently...
# ---------------------------------------------------------------

set request [ns_conn request]
if {![regexp {action=([a-z]+).*?project_id=([0-9]+)} $request match action project_id]} {
    set action ""
    set project_id ""
    ad_return_complaint 1 "budget-tree-action: not recognized: request=$request"
    ad_script_abort
}

set current_user_id [auth::require_login]
set current_url [im_url_with_query]
set user_admin_p [im_is_user_site_wide_or_intranet_admin $current_user_id]

ns_log Notice "budget-tree-action: form_vars: action=$action, project_id=$project_id, current_user_id=$current_user_id, user_admin_p=$user_admin_p"

# ---------------------------------------------------------------
# 
# ---------------------------------------------------------------

ad_proc im_budget_tree_action {
    -action
    -json_list
} {
    Process an update on budget or cost items and update the database
    accordingly
} {
    ns_log Notice "action -action=$action, -json_list=$json_list"
    set current_user_id [auth::require_login]
    
    # Get hash with var-value entries from JSON
    array set json_hash $json_list
    set object_id $json_hash(id)
    if {"" eq $object_id} { return }; # May happen when trying to save incomplete model
    set object_type $json_hash(object_type)
    if {"" eq $object_type} { 
	ns_log Error "action: found empty object_type='$object_type': json_list=$json_list"
	return 
    }
    ns_log Notice "action: After getting oid and otype: object_id=$object_id, object_type=$object_type"
    if {$object_type ni {"im_budget_item" "im_cost" "im_invoice" "im_timesheet_invoice"}} {
	ns_log Error "action: unrecognized object_type=$object_type, skipping"
	return
    }

    # ----------------------------------------------------------
    # Permissions
    # ----------------------------------------------------------
    
    ${object_type}_permissions $current_user_id $object_id view_p read_p write_p admin_p
    if {!$read_p} {
	set success "false"
	set message "Insufficient permissions to read item"
	im_rest_error -format json -http_status 403 -message $message
	return
    }
    switch $action {
	update {
	    if {!$write_p} {
		set success "false"
		set message "Insufficient permissions to modify item"
		im_rest_error -format json -http_status 403 -message $message
		return
	    }
	}
	delete {
	    if {!$admin_p} {
		set success "false"
		set message "Insufficient permissions to delete item"
		if {"im_budget_item" ne $object_type} { set message "You can't delete this item" }
		im_rest_error -format json -http_status 403 -message $message
		return
	    }
	}
    }


    # ----------------------------------------------------------
    # Delete action
    # ----------------------------------------------------------
    if {"delete" eq $action} {

	db_string del "select im_budget_item__delete(:object_id)"
	return

    }



    # ----------------------------------------------------------
    # Update action
    # ----------------------------------------------------------

    # Check if the object exists and create
    # The JSON includes a valid object_id that came from the acs_objects_seq before
    set exists_p [db_string exists "select count(*) from im_budget_items where budget_item_id = :object_id"]
    if {!$exists_p && "im_budget_item" eq $object_type} {
	# Create the object
	
	set vars {budget_item_name budget_item_nr budget_item_parent_id budget_item_type_id budget_item_status_id budget_item_project_id}
	foreach var $vars {
	    set $var $json_hash($var)
	}
	if {"root" eq $budget_item_parent_id} { set budget_item_parent_id "" }

	set id [db_string new_bi "select im_budget_item__new (
		:object_id,
		'im_budget_item',
		now(),
		:current_user_id,
		'[ns_conn peeraddr]',
		null,

		:budget_item_name,
		:budget_item_nr,
		:budget_item_parent_id,
		:budget_item_type_id,
		:budget_item_status_id,
		:budget_item_project_id
	)"]
	ns_log Notice "action: Created new BI with budget_item_id=$id with id=$id from object_id=$object_id"
	if {$id ne $object_id} { 
	    ns_log Error "action: After create new BI: New id=$id is different from object_id=$object_id" 
	}
    }

    # Get hash with var-value entries from database
    set audit_value [im_audit_object_value -object_id $object_id]
    set audit_lines [split $audit_value "\n"]
    foreach line $audit_lines {
	set line_pieces [split $line "\t"]
	set var [lindex $line_pieces 0]
	set val [lindex $line_pieces 1]
	if {"" eq $var} { continue }
	set audit_hash($var) $val
	# ns_log Notice "action: audit_hash('$var') = '$val'"
    }

    set ignore_vars {
creation_date
creation_ip
creation_user
expanded
id
last_modified
leaf
modifying_ip
modifying_user
name
max_child_sortkey
nr
object_type
parentId
tree_sortkey
    }

    # Translate between JSON variable names and database column names
    switch $object_type {
	"im_budget_item" {
	    set translate_hash(name) budget_item_name
	    set translate_hash(nr) budget_item_nr
	    set translate_hash(object_status_id) budget_item_status_id
	    set translate_hash(object_type_id) budget_item_type_id
	}
	"im_cost" - "im_invoice" - "im_timesheet_invoice" {
	    set translate_hash(object_status_id) cost_status_id
	    set translate_hash(object_type_id) cost_type_id
	    set translate_hash(description) note
	}
	default {
	    ns_log Error "action: translate: unrecognized object_type='$object_type'"
	}
    }

    set audit_vars [array names audit_hash]
    # ns_log Notice "action: audit_vars=$audit_vars"

    foreach var [array names json_hash] {

	if {$var in $ignore_vars} { continue }
	if {[regexp {^cost_type_} $var match]} { continue }

	# Translate varnames
	set audit_var $var
	set json_var $var
	if {[info exists translate_hash($audit_var)]} { 
	    # ns_log Notice "action: found translation_hash($audit_var) = $translate_hash($audit_var)"
	    set audit_var $translate_hash($audit_var) 
	}

	# Get values
	set json_val $json_hash($json_var)
	if {"budget_item_parent_id" eq $json_var && "root" eq $json_val } { set json_val "" }; # JSON "root" -> null

	if {"" ne $json_val && $audit_var ni $audit_vars} {
	    # Check if the json var exists in the DB. Ignore empty values...
	    ns_log Error "action: found json_var='$json_var', audit_var=$audit_var, val='$json_val' which doesn't exist in audit_vars=$audit_vars"
	    continue
	}
	set audit_val ""
	if {[info exists audit_hash($audit_var)]} { set audit_val $audit_hash($audit_var) }


	# Check for need to update
	if {$audit_val eq $json_val} { continue }


	# -----------------------------------------------------
	# Update the value
	ns_log Notice "action: Found update: json_var=$json_var, audit_var=$audit_var, audit_val=$audit_val, json_val=$json_val"

	set id $json_hash(id)
	set object_type $json_hash(object_type)
	ns_log Notice "action: id=$object_id, object_type=$object_type"

	switch $object_type {
	    "im_budget_item" {
		ns_log Notice "action: update im_budget_items set $audit_var = $json_val where budget_item_id = $object_id"
		db_dml update_bi "update im_budget_items set $audit_var = :json_val where budget_item_id = :object_id"
	    }
	    "im_cost" - "im_invoice" - "im_timesheet_invoice" {
		# Only allow to update the budget_item_id. Everything else is for the owner of the cost item
		set cost_vars {budget_item_id}
		if {$audit_var ni $cost_vars} { continue }
		ns_log Notice "action: update costs set $audit_var = '$json_val' where budget_item_id = '$object_id'"
#		db_dml update_bi "update costs set $audit_var = :json_val where budget_item_id = :object_id"
	    }
	    default {
		ns_log Error "action: Unknown object_type=$object_type"
	    }
	}

    }

    ns_log Notice "action: Finished"

    return 0; # Return 1 if we want to iterate
}


# ---------------------------------------------------------------
# 
# ---------------------------------------------------------------

# Default values for JSON return message
set success "true"
set message "Successfully performed action=$action"

if {[catch {

    # Parse the JSON POST data
    set post_content [ns_conn content]
    array set json_hash [util::json::parse $post_content]
    # ns_log Notice "budget-tree-action: json_hash=[array get json_hash]"
    
    # ---------------------------------------------------------------
    # Check for single update
    # ---------------------------------------------------------------
    
    if {[info exists json_hash(_object_)]} {
	set json_list $json_hash(_object_)
	# ns_log Notice "budget-tree-action: object: json_list=$json_list"
	im_budget_tree_action -action $action -json_list $json_list
    }

    # ---------------------------------------------------------------
    # Check for multiple updates
    # ---------------------------------------------------------------
    
    if {[info exists json_hash(_array_)]} {
	set json_array $json_hash(_array_)
	
	# ns_log Notice "budget-tree-action: array=$json_array"
	set repeat_p 1
	set cnt 0
	while {$repeat_p && $cnt < 100} {
	    set repeat_p 0
	    incr cnt
	    foreach array_elem $json_array {
		ns_log Notice "budget-tree-action: cnt=$cnt, array_elem=$array_elem"
		set obj [lindex $array_elem 0]
		set json_list [lindex $array_elem 1]
		ns_log Notice "budget-tree-action: cnt=$cnt, decomposing array_elem: $obj=$json_list"
		set not_finished_p [im_budget_tree_action -action $action -json_list $json_list]
		if {1 eq $not_finished_p} { set repeat_p 1 }
	    }
	}
    }
} err_msg]} {
    ns_log Error "budget-tree-action: Reporting back error: [ad_print_stack_trace]"
    set success "false"
    set message $err_msg
    im_rest_error -format json -http_status 404 -message "Internal Error: [ad_print_stack_trace]"

}

