# /packages/intranet-budget/www/budget-tree.json.tcl
#
# Copyright (C) 2013-2020 ]project-open[

ad_page_contract {
    Returns a JSON tree structure suitable for batch-loading a budget TreeStore
    @param project_id Top project (if applies)
    @author frank.bergmann@budget-open.com
    
    @param node Passed by ExtJS to load sub-trees of a tree.
                Normally not used, just in case of error.
} {
    project_id:integer
    {debug_p 0}
    {node ""}
}


set main_project_id $project_id
set root_project_id $project_id
if {"" ne $node && [string is integer $node]} { set root_project_id $node }

db_1row project_info "
	select	project_name,
		project_nr
	from	im_projects
	where	project_id = :main_project_id
"
ns_log Notice "budget-task-tree.json: node=$node, main_project_id=$main_project_id, root_project_id=$root_project_id"


# --------------------------------------------
# Security & Permissions
#
set current_user_id [auth::require_login]
im_project_permissions $current_user_id $main_project_id view read write admin
if {!$read} {
    im_rest_error -format "json" -http_status 403 -message "You (user #$current_user_id) have no permissions to read project #$main_project_id"
    ad_script_abort
}


# --------------------------------------------
# Make sure there is atleast one BI
#
set top_project_budget_item_id [db_string main_bi "
	select	min(bi.budget_item_id)
	from	im_budget_items bi
	where	bi.budget_item_project_id in (
			select	main_p.project_id
			from	im_projects main_p,
				im_projects p
			where	p.project_id = :main_project_id and
				main_p.tree_sortkey = tree_root_key(p.tree_sortkey)
		) and
		bi.budget_item_parent_id is null
"]

if {"" eq $top_project_budget_item_id} {
    # Create a new top default budget item
    set top_project_budget_item_id [db_string top_bi "
	select im_budget_item__new(
		null, 'im_budget_item', now(), 0, '0.0.0.0', null,
		'Top Budget Item for ' || :project_name,
		'budget_item_' || :project_nr,
		null, 49100, 49000, :main_project_id
	)
    "]
}


# --------------------------------------------
# Get the list of budget items related to project
# Budget items are associated to the main project only.
# Make the list unique and sort by tree_sortkey, so that the top item appears first.
set project_bis [db_list bis "
	select	sort_bi.budget_item_id
	from	(select distinct
			ref_bi.budget_item_id
		from	(
				select	bi.budget_item_id
				from	im_budget_items bi
				where	bi.budget_item_project_id = :main_project_id
			UNION
				select	bi.budget_item_id
				from	im_projects main_p,
					acs_rels r,
					im_budget_item_project_rels bipr,
					im_budget_items bi
				where	main_p.project_id = :main_project_id and
					r.object_id_one = main_p.project_id and
					r.object_id_two = bi.budget_item_id and
					r.rel_id = bipr.rel_id
			) t,
			im_budget_items ref_bi
		where	t.budget_item_id = ref_bi.budget_item_id
		order by ref_bi.budget_item_id
		) tt,
		im_budget_items sort_bi
	where	tt.budget_item_id = sort_bi.budget_item_id
	order by sort_bi.tree_sortkey
"]
lappend project_bis 0
set main_project_budget_item_id [lindex $project_bis 0]
# ad_return_complaint 1 "budget-tree.json.tcl: project_bis=$project_bis<br>main_project_id=$main_project_id<br>main_project_bi=$main_project_budget_item_id"


# --------------------------------------------
# Get all the variables valid for budget item
#
set valid_vars1 [util_memoize [list im_rest_object_type_columns -deref_p 0 -rest_otype "im_budget_item"]]
set valid_vars2 [util_memoize [list im_rest_object_type_columns -deref_p 0 -rest_otype "im_invoices"]]
set valid_vars3 {id name nr parent_id level object_type_id object_status_id num_children sort_order expanded}
set valid_vars [set_union [set_union $valid_vars1 $valid_vars2] $valid_vars3]

set rollup_cost_type_ids [im_budget_item_rollup_cost_type_ids]
foreach cost_type_id $rollup_cost_type_ids { lappend valid_vars cost_type_$cost_type_id }

set valid_vars [lsort -unique $valid_vars]
# ad_return_complaint 1 "budget-tree.json.tcl: <pre>valid_vars=<br>[join $valid_vars "\n"]</pre>"




# --------------------------------------------
# Main hierarchical SQL
#
set budgets_sql "
	select distinct
		bi.budget_item_id as id,
		bi.budget_item_name as name,
		bi.budget_item_nr as nr,
		bi.budget_item_parent_id as parent_id,
		tree_level(bi.tree_sortkey) as level,
		coalesce(bi.sort_order,0) as bi_sort_order,
		CASE WHEN bts.open_p = 'o' THEN 'true' ELSE 'false' END as expanded,
		bi.budget_item_type_id as object_type_id,
		bi.budget_item_status_id as object_status_id,
		bi.*,
		o.*,
		bo.*
	from	im_budget_items main_bi,
		im_budget_items bi
		LEFT OUTER JOIN acs_objects o ON (bi.budget_item_id = o.object_id)
		LEFT OUTER JOIN im_biz_objects bo ON (bi.budget_item_id = bo.object_id)
		LEFT OUTER JOIN im_biz_object_tree_status bts ON (
			bi.budget_item_id = bts.object_id and 
			bts.page_url = 'default' and
			bts.user_id = :current_user_id
		)
	where	main_bi.budget_item_id in ([join $project_bis ","]) and
		bi.tree_sortkey between main_bi.tree_sortkey and tree_right(main_bi.tree_sortkey) and
		bi.budget_item_status_id not in (select im_sub_categories([im_budget_item_status_deleted]))
	order by
		level,
		coalesce(bi.sort_order, 0),
		bi.tree_sortkey
"
# ad_return_complaint 1 "budget-tree.json.tcl:<br><pre>$budgets_sql</pre><br>[im_ad_hoc_query -format html $budgets_sql]"


# ---------------------------------------------------
# Write budget items into 3 hashes
#
set object_top_count 0
db_foreach budgets $budgets_sql {
    # Store the object type of each oid
    set object_type_hash($budget_item_id) $object_type
    set object_level_hash($budget_item_id) $level

    # Write the object value into a hash
    if {"" == $parent_id} { set expanded "true" }
    array unset hash
    foreach v $valid_vars {
	set hash($v) [im_opt_val -limit_to nohtml $v]
    }
    set object_var_hash($budget_item_id) [array get hash]
    # ad_return_complaint 1 [array get hash]

    # Create a hash of children to support fast recursive traversal
    set object_parent_hash($budget_item_id) $parent_id
    set children [list]
    if {[info exists object_children_hash($parent_id)]} { set children $object_children_hash($parent_id) }
    lappend children $budget_item_id
    set object_children_hash($parent_id) $children

    # Collect the list of top elements
    if {"" eq $parent_id} {
	set object_top_hash($object_top_count) $budget_item_id
	incr object_top_count
    }
}

foreach v $valid_vars { set $v "" }



# ---------------------------------------------------
# Add Financial documents below Budget Items
#

set findoc_sql "
	select	c.cost_id as id,
		c.cost_name as name,
		c.cost_nr as nr,
		coalesce(c.budget_item_id, :main_project_budget_item_id) as parent_id,
		c.cost_type_id as object_type_id,
		c.cost_status_id as object_status_id,

		round(coalesce(c.amount, 0.0),3) as amount,
		c.cost_center_id,
		c.cost_id,
		c.cost_name,
		c.cost_nr,
		c.cost_status_id,
		c.cost_type_id,
		c.currency,
		c.customer_id,
		c.description,
		c.effective_date,
		c.note,
		c.project_id,
		c.provider_id,

		o.*,
		i.*,
		'false' as expanded
	from	acs_objects o,
		im_costs c
		LEFT OUTER JOIN im_invoices i ON (c.cost_id = i.invoice_id)
	where	c.cost_id = o.object_id and
		c.cost_type_id not in (select im_sub_categories(3708)) and
		c.project_id in (
			select	sub_p.project_id
			from	im_projects sub_p,
				im_projects main_p
			where	main_p.project_id = :project_id and
				sub_p.tree_sortkey between main_p.tree_sortkey and tree_right(main_p.tree_sortkey)
	       )
	order by
		c.cost_type_id,
		c.cost_nr
"
# ad_return_complaint 1 "budget-tree.json.tcl:<br><pre>$findoc_sql</pre><br>[im_ad_hoc_query -format html $findoc_sql]"

# ---------------------------------------------------
# Write costs into 3 hashes
#
db_foreach findocs $findoc_sql {

    if {$cost_type_id eq 3718 && $cost_id > 40000} { continue }
    if {$cost_type_id eq 3718 && $cost_name eq "2015_05_10 System Administrator"} { continue }


    # Store the object type of each oid
    set object_type_hash($cost_id) $object_type
    set object_level_hash($cost_id) 0

    # Calculate new level
    set parent_level 0
    if {[info exists object_level_hash($parent_id)]} { set parent_level $object_level_hash($parent_id) }
    set level [expr $parent_level + 1]
    set object_level_hash($cost_id) $level

    # Write the object value into a hash
    array unset hash
    foreach v $valid_vars {
	set hash($v) [im_opt_val -limit_to nohtml $v]
    }
    set object_var_hash($cost_id) [array get hash]
    # ad_return_complaint 1 [array get hash]

    # Create a hash of children to support fast recursive traversal
    set object_parent_hash($cost_id) $parent_id
    set children [list]
    if {[info exists object_children_hash($parent_id)]} { set children $object_children_hash($parent_id) }
    lappend children $cost_id
    set object_children_hash($parent_id) $children

    # Seed the aggregation of cost_types
    set object_rollup_hash($cost_id) [list $cost_type_id $amount]
}
# ad_return_complaint 1 "budget-tree.json.tcl: object_rollup_hash:<br><pre>[array get object_rollup_hash]</pre>"


# ---------------------------------------------------
# Aggregate / Roll-Up values from costs up to the top BI
# This procedure assumes that leaf objects have an entry
# for object_rollup_hash.
#
set debug {}
foreach cost_id [array names object_rollup_hash] {

    set cost_hash_list $object_rollup_hash($cost_id)
    set cost_type_id [lindex $cost_hash_list 0]
    set amount [lindex $cost_hash_list 1]

    ns_log Notice debug "budget-tree.json.tcl: cost_id=$cost_id, cost_type_id=$cost_type_id, amount=$amount"
    lappend debug "cost_id=$cost_id, cost_type_id=$cost_type_id, amount=$amount"

    set cnt 0
    set oid $object_parent_hash($cost_id)
    while {"" ne $oid && $cnt < 100} {
	
	set rollup_hash_list [list]
	if {[info exists object_rollup_hash($oid)]} { set rollup_hash_list $object_rollup_hash($oid) }
	array unset rollup_hash
	array set rollup_hash $rollup_hash_list
	
	set v 0.0
	if {[info exists rollup_hash($cost_type_id)]} { set v $rollup_hash($cost_type_id) }
	set v [expr round(100.0 * ($v + $amount)) / 100.0]
	set rollup_hash($cost_type_id) $v
	
	# Write updated hash into parent
	set object_rollup_hash($oid) [array get rollup_hash]

	ns_log Notice debug "budget-tree.json.tcl: oid=$oid, hash=[array get rollup_hash]"
	lappend debug "oid=$oid, hash=[array get rollup_hash]"

	array unset rollup_hash

	# Advanced to next level top
	incr cnt
	set oid $object_parent_hash($oid)
    }
}

# ad_return_complaint 1 "<pre>[join $debug "\n"]</pre>"


# ---------------------------------------------------
# Write out JSON using recursive depth-first descend
#
set debug_list ""
set stack [list]
foreach top_ctr [array names object_top_hash] { lappend stack $object_top_hash($top_ctr) }
lappend debug_list "Start: stack=$stack"
set ctr 0
set old_level 1
set budget_json ""
set indent ""

while {[llength $stack] > 0 && $ctr < 100} {
    lappend debug_list "ctr=$ctr: stack=$stack"

    # -----------------------------------------
    # Remove the first element from the stack
    set tos [lindex $stack 0]
    set stack [lrange $stack 1 end]
    lappend debug_list "ctr=$ctr: tos=$tos, stack=$stack"
    
    # Add children to the stack
    set children [list]
    if {[info exists object_children_hash($tos)]} { set children $object_children_hash($tos) }
    lappend debug_list "ctr=$ctr: tos=$tos, children=$children"
    set stack [concat $children $stack]

    # -----------------------------------------
    # Load the values for the item
    set hash_vals $object_var_hash($tos)
    array unset hash
    array set hash $hash_vals
    foreach v $valid_vars {
	set $v $hash($v)
    }


    # -----------------------------------------
    # Close off the previous entry
    lappend debug_list "ctr=$ctr: finish: level=$level, old_level=$old_level\n"

    # This is the first child of the previous item. Increasing the level always happens in steps of 1
    if {$level > $old_level} { append budget_json ",\n${indent}\tchildren:\[\n" }
    
    # A group of children needs to be closed. Please note that this can cascade down to several levels.
    while {$level < $old_level} {
	append budget_json "\n${indent}\}\]\n"
	incr old_level -1
	set indent ""
	for {set i 0} {$i < $old_level} {incr i} { append indent "\t" }
    }

    # The current task is on the same level as the previous. This is also executed after reducing the old_level in the previous while loop
    if {$level == $old_level} {
	if {0 != $ctr} { append budget_json "${indent}\n${indent}\},\n"	}
    }

    lappend debug_list "ctr=$ctr: bi=$budget_item_name, level=$level, old_level=$old_level\n"

    set indent ""
    for {set i 0} {$i < $level} {incr i} { append indent "\t" }
    
    set quoted_char_map {"\n" "\\n" "\r" "\\r" "\"" "\\\"" "\\" "\\\\"}
    set quoted_budget_item_name [string map $quoted_char_map $budget_item_name]

    # ToDo: Remove other icon types and deal with different object types to be displayed
    set icon_type "undefined"
    switch $object_type_id {
    	49100 { set icon_type "budget-item" }
    	3700 { set icon_type "invoice" }
    	3702 { set icon_type "quote" }
    	3704 { set icon_type "provider-bill" }
    	3706 { set icon_type "purchase-order" }
    	3718 { set icon_type "timesheet-cost" }
    	3720 { set icon_type "expense-item" }
    	3722 { set icon_type "expense-bundle" }
    	3724 { set icon_type "delivery-note" }
	default {
	    set icon_type "error"
	    ns_log Error "budget-tree.json.tcl: Unknown object_type_id=$object_type_id ([im_category_from_id $object_type_id])"
	}
    }
    
    append budget_json "${indent}\{\n${indent}\ticonCls:\"icon-$icon_type\",\n"
    
    set rollup_hash_list {}
    if {[info exists object_rollup_hash($tos)]} { set rollup_hash_list $object_rollup_hash($tos) }
    array unset rollup_hash
    array set rollup_hash $rollup_hash_list
    foreach cost_type_id $rollup_cost_type_ids {
	set v ""
	if {[info exists rollup_hash($cost_type_id)]} { set v $rollup_hash($cost_type_id) }
	if {"" ne $v} {
	    append budget_json "${indent}\tcost_type_$cost_type_id: $v,\n"
	}
    }

    #ad_return_complaint 1 $valid_vars
    foreach var $valid_vars {
	# Skip xml_* variables (only used by MS-Budget)
	if {[regexp {^xml_} $var match]} { continue }
	if {[regexp {^creation_} $var match]} { continue }
	if {[regexp {^modifying_} $var match]} { continue }
	if {[regexp {^last_} $var match]} { continue }
	if {[regexp {^title} $var match]} { continue }
	if {[regexp {^security_inherit_p} $var match]} { continue }

	# Append the value to the JSON output
	set value [set $var]
	if {"" ne $value} {
	    set quoted_value [string map $quoted_char_map $value]
	    append budget_json "${indent}\t$var:\"$quoted_value\",\n"
	}
    }
    # ad_return_complaint 1 $budget_json


    # Add information about leaf or not leaf (children)
    set children [list]
    if {[info exists object_children_hash($tos)]} { set children $object_children_hash($tos) }
    set num_children [llength $children]
    if {0 == $num_children} { set leaf_json "true" } else { set leaf_json "false" }
    append budget_json "${indent}\tleaf:$leaf_json"

    incr ctr
    set old_level $level
}

set level 0
while {$level < $old_level} {
    # A group of children needs to be closed.
    # Please note that this can cascade down to several levels.
    append budget_json "\n${indent}\}\]\n"
    incr old_level -1
    set indent ""
    for {set i 0} {$i < $old_level} {incr i} { append indent "\t" }
}

set json "{'text':'.','children': \[
$budget_json
}
"

if {0} {
    ad_return_complaint 1 "<pre>
 object_chidren_hash:\n[join [array get object_children_hash] "\n"]\n\n
 object_top_hash:\n[array get object_top_hash]\n\n[array get object_level_hash]\n\n
 object_var_hash:\n[join $debug_list "\n"]\n\n[join [array get object_var_hash] "\n"]\n\n
 json:\n$json</pre>"
}

